Set Implicit Arguments.

Require Import AdditionalTactics.
Require Import  List Decidable Metatheory.
Require Import Core.
Require Export Wf_nat.
Require Import Omega.

Inductive term : tm -> Prop :=
  | term_fv  : forall a,  term (tm_fv a)
  | term_con : forall a,  term (tm_con a)
  | term_ctr : forall a,  term (tm_ctr a)
  | term_app : forall t1 t2,   
    term t1 -> 
    term t2 ->
    (~exists v, t1 = tm_app P_SAYS v) ->
    (~exists v, t1 = tm_app PF_SIGN v)->
    t1 <>T_PF ->
    t1 <> TM_PF_RETURN ->
    (~exists v, t1 = tm_app TM_PF_BIND v)->
    t1 <> TM_SAY ->
     (~exists v1, exists v2, exists v3, t1 = tm_app( tm_app  (tm_app TM_PIF v1) v2) v3) ->
    (~exists v, t1 = tm_app PF_SAYS_RETURN v) ->
     (~exists v, t1 = tm_app  PF_SAYS_BIND v) ->
    term (tm_app t1 t2)
  | term_abs : forall L t1 t2, term t1 -> 
                (forall x, x \notin L -> term (open_tm (tm_fv x) t2)) ->
                term (tm_abs t1 t2)
  | term_arr : forall L t1 t2, term t1 -> 
                (forall x, x \notin L -> term (open_tm (tm_fv x) t2)) ->
                term (tm_arr t1 t2)
  | term_mat : forall t1 t2 t3, term t1 -> term t2 -> term t3 -> term (tm_mat t1 t2 t3)
  | term_brn : forall c t2 t3,  term t2 -> term t3 -> term (tm_brn c t2 t3)
  | term_cast : forall t1 t2,  term t1 -> term t2 -> term (tm_cast t1 t2)
  | term_P_SAYS: forall t1 t2, term t1 -> term t2 ->
                 term (tm_app (tm_app P_SAYS t1) t2)
  | term_PF_SIGN: forall t1 t2, term t1 -> term t2 ->
                 term (tm_app (tm_app PF_SIGN t1) t2)
  | term_T_PF: forall t, term t ->
                 term (tm_app T_PF t)
  | term_TM_PF_RETURN: forall t, term t ->
                 term (tm_app TM_PF_RETURN t)
  | term_TM_PF_BIND: forall t1 t2, term t1 -> term t2 ->
                 term (tm_app (tm_app TM_PF_BIND t1) t2)
  | term_TM_SAY: forall t, term t ->
                 term (tm_app TM_SAY t)
  | term_PIF : forall t1 t2 t3 t4, term t1 -> term t2 -> term t3 -> term t4 -> 
               term (tm_app (tm_app (tm_app (tm_app TM_PIF t1) t2) t3) t4)
  | term_PF_SAYS_RETURN: forall t1 t2, term t1 -> term t2 ->
                 term (tm_app (tm_app PF_SAYS_RETURN t1) t2)
  | term_PF_SAYS_BIND: forall t1 t2, term t1 -> term t2 ->
                 term (tm_app (tm_app PF_SAYS_BIND t1) t2)
 .

Fixpoint tm_size (t:tm) {struct t} : nat :=
  match t with
  | tm_fv _         => 1
  | tm_bv n         => 1
  | tm_con _        => 1
  | tm_ctr _        => 1
  | tm_app t1 t2    => 1 + (tm_size t1) + (tm_size t2)
  | tm_abs t1 t2    => 1 + (tm_size t1) + (tm_size t2)
  | tm_arr t1 t2    => 1 + (tm_size t1) + (tm_size t2) 
  | tm_mat t1 t2 t3  => 1 + (tm_size t1) + (tm_size t2)  + (tm_size t3) 
  | tm_brn c t2 t3  => 2 + (tm_size t2) + (tm_size t3) 
  | tm_cast t1 t2   => 1 + (tm_size t1) + (tm_size t2) 
  end
  .

Lemma open_tm_size_eq : forall k X T,
  tm_size (open_tm_aux k (tm_fv X) T) = tm_size T.
Proof.
  intros k X T.
  generalize dependent k.
  generalize dependent X.
  induction T; intros X k; simpl; auto.
    destruct (b===k); simpl; reflexivity.
Qed.

(***********************************************************)
(** binds dec *)

Lemma __binds_var_dec : forall E a t,
  (forall a1 a2:bnd, {a1 = a2} + {~ a1 = a2}) ->
  decidable (binds_var a t E).
Proof.
  introv Dec. remember (get a E) as B. symmetry in HeqB.
  unfold binds_var. destruct B as [a'|].
  destruct (Dec (bnd_var t) a'). subst. 
    left*.
    right. intros H. congruence.
  right. intros H. congruence.
Qed.

Lemma _binds_var_dec : forall E a t,
  decidable (binds_var a t E).
Proof.
  intros. apply __binds_var_dec. apply eq_bnd_dec.
Qed.

Lemma binds_var_dec: forall E a,
   decidable (exists t, binds_var a t E).
Proof. 
   induction E; intro; unfold binds_var; unfold binds; unfold decidable; simpl.
     right. unfold not. intros. destruct H. inversion H.
     destruct a. destruct b. 
       destruct (a0==a).
         left. apply ex_intro with (x:=t). auto.
         destruct (@IHE a0).
           destruct H.  left. apply ex_intro with (x:=x). unfold binds_var in H. unfold binds in H. auto. 
           unfold binds_var in H. unfold binds in H.
             right. auto.
       destruct (a0==a).
         right. unfold not. intros. destruct H. inversion H.
         destruct (@IHE a0).
           destruct H.  left. apply ex_intro with (x:=x). unfold binds_var in H. unfold binds in H. auto. 
           unfold binds_var in H. unfold binds in H.
             right. auto.
Qed.

(***********************************************************)
Lemma is_tm_arr_dec: forall k1 k2, 
  decidable (exists u, k1 = tm_arr k2 u).
Proof. 
  intros. unfold decidable. 
  destruct k1; try solve [right; unfold not; intro; destruct H as [u H]; inversion H].
    Case "tm_arr".
    assert ({k1_1=k2}+{~k1_1= k2}).
      apply eq_tm_dec.
    destruct H.
      left. subst. apply ex_intro with (x:=k1_2). reflexivity.
      right. unfold not. intros. destruct H as [u H]. inversion H. unfold not in n. contradiction.
Qed.

Lemma is_tm_app_dec: forall t,
  decidable (exists u1, exists u2, t = tm_app u1 u2).
Proof. 
  intros. unfold decidable. 
  destruct t;
    try solve [right; unfold not; intro; destruct H as [u1 H]; destruct H as [u2 H]; inversion H].
    left. apply ex_intro with (x:=t1). apply ex_intro with (x:=t2). reflexivity.
Qed.

Lemma is_tm_app_dec2: forall t u1,
  decidable (exists u2, t = tm_app u1 u2).
Proof. 
  intros. unfold decidable. 
  destruct t;
    try solve [right; unfold not; intro; destruct H as [u2 H]; inversion H].
    assert ({t1=u1}+{~t1= u1}).
      apply eq_tm_dec.
    destruct H.
      left. subst. apply ex_intro with (x:=t2). reflexivity.
      right; unfold not; intro; destruct H as [u2 H]; inversion H; contradiction.
Qed.

Lemma is_tm_app_dec3: forall t u1,
  decidable (exists v1 : tm,
                      exists v2 : tm,
                        exists v3 : tm,
                          t = tm_app (tm_app (tm_app u1 v1) v2) v3).
intros. unfold decidable. 
  destruct t;
    try solve [right; unfold not; intro; destruct H as [v3 H]; destruct H as [v2 H]; destruct H as [v1 H]; inversion H].
  destruct t1;
    try solve [right; unfold not; intro; destruct H as [v3 H]; destruct H as [v2 H]; destruct H as [v1 H]; inversion H].
  destruct t1_1;
    try solve [right; unfold not; intro; destruct H as [v3 H]; destruct H as [v2 H]; destruct H as [v1 H]; inversion H].
  assert ({t1_1_1=u1}+{~t1_1_1= u1}).
      apply eq_tm_dec.
  destruct H.
      left. subst. 
        apply ex_intro with (x:=t1_1_2). 
        apply ex_intro with (x:=t1_2). 
        apply ex_intro with (x:=t2). 
        reflexivity.
      right; unfold not; intro; destruct H as [v1 H]; destruct H as [v2 H]; destruct H as [v3 H]; inversion H; contradiction.
Qed.

(*********************************************************)
(** value applied_value dec *)
Lemma value__applied_value__appPF_SIGN_dec: forall t, 
  decidable (value t) /\ 
  decidable (applied_value t) /\ 
  decidable (exists u2, t = tm_app PF_SIGN u2 /\ value u2)
  .
Proof.
  intros. unfold decidable. 
  induction t; split; 
    try solve [ split;
      try solve [
          right; unfold not; intro; inversion H
        | right; unfold not; intro; destruct H as [u2 H]; inversion H; inversion H0
        ]
      ].
    Case "fv v".
    left. apply value_var.
    Case "bv v".
    right. unfold not. intro. inversion H. inversion H0.
    Case "con v".
    destruct (01===c).
      (*S_KIND := tm_con 01*)
      left. subst. apply value_KIND.
    destruct (10===c).
      (*K_TYPE := tm_con 10*)
      left. subst. apply value_TYPE.
    destruct (11===c).
      (*K_PROP := tm_con 11*)
      left. subst. apply value_PROP.
    destruct (30===c).
      (*T_PRIN := tm_con 30*)
      left. subst. apply value_T_PRIN.
    destruct (44===c).
      (*TM_SELF:= tm_con 44*)
      left. subst. apply value_self.
    unfold not in *. right. unfold not. intro. inversion H; try solve [contradiction].
      inversion H0.
    Case "ctr v".
    left. apply value_app. apply applied_value_ctr.
  split.
    Case "ctr av".
    left. apply applied_value_ctr.
    Case "ctr ex".
    right; unfold not; intro; destruct H as [u2 H]; inversion H; inversion H0.
    Case "app v".
    decompose [and] IHt1. decompose [and] IHt2. clear IHt1. clear IHt2.
      destruct H; destruct H1; destruct H2; destruct H0; destruct H4; destruct H5;
       try solve [
        left; apply value_app; apply applied_value_app; auto
      | right; unfold not; intro; inversion H5; subst;
         try solve [
             inversion H6; contradiction
           | inversion H1;subst; inversion H9
           | inversion H1;subst; inversion H8
           | inversion H1;subst; inversion H1
          ]
     | right; unfold not; intro; inversion H5; subst;
          try solve[
            inversion H6; contradiction
          | inversion H;subst; inversion H6;subst; inversion H10
          | inversion H;subst; inversion H6;subst; inversion H9
          | inversion H;subst; inversion H6;subst
          ]
     | destruct H2 as [u2 H2]; destruct H2; subst;
       left; apply value_sign; auto
      ].
     (** t1 isnt v, t1 isnt av, t1 does not ex, t2 is v, t2 is av, t2 does ex *)
     assert ({t1=T_PF}+{~t1=T_PF}). 
       apply eq_tm_dec.
      destruct H5. left. subst. apply value_T_PF.

      assert (decidable (exists u, t1 = tm_app P_SAYS u)).
       apply is_tm_app_dec2. 
      destruct H5.
       destruct H5 as [u H5]. subst.
       left. subst. apply value_P_SAYS; auto.

     assert ({t1=TM_PF_RETURN}+{~t1=TM_PF_RETURN}).
       apply eq_tm_dec.
      destruct H6. left. subst. apply value_TM_PF_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_RETURN u)).
       apply is_tm_app_dec2. 
      destruct H6.
       destruct H6 as [u H6]. subst.
       left. subst. apply value_PF_SAYS_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_BIND u)).
       apply is_tm_app_dec2. 
      destruct H7.
       destruct H7 as [u H7]. subst.
       left. subst. apply value_PF_SAYS_BIND; auto.

       right; unfold not; intro; inversion H8; subst.
          inversion H9; contradiction.
          inversion H8; subst. 
            inversion H9;subst. inversion H13; subst. inversion H15.
            unfold not in H2. apply H2. apply ex_intro with (x:=v). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H5. apply H5. apply ex_intro with (x:=v). auto.
          unfold not in n. apply n. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H13; subst.
            unfold not in n0. apply n0. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H6. apply H6. apply ex_intro with (x:=a). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H7. apply H7. apply ex_intro with (x:=e1). auto.
     (** t1 isnt v, t1 isnt av, t1 does not ex, t2 is v, t2 is av, t2 doesnt ex *)
     assert ({t1=T_PF}+{~t1=T_PF}).
       apply eq_tm_dec.
     destruct H5. left. subst. apply value_T_PF.

      assert (decidable (exists u, t1 = tm_app P_SAYS u)).
       apply is_tm_app_dec2. 
      destruct H5.
       destruct H5 as [u H5]. subst.
       left. subst. apply value_P_SAYS; auto.

     assert ({t1=TM_PF_RETURN}+{~t1=TM_PF_RETURN}).
       apply eq_tm_dec.
      destruct H6. left. subst. apply value_TM_PF_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_RETURN u)).
       apply is_tm_app_dec2. 
      destruct H6.
       destruct H6 as [u H6]. subst.
       left. subst. apply value_PF_SAYS_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_BIND u)).
       apply is_tm_app_dec2. 
      destruct H7.
       destruct H7 as [u H7]. subst.
       left. subst. apply value_PF_SAYS_BIND; auto.

       right; unfold not; intro; inversion H8; subst.
          inversion H9; contradiction.
          inversion H8; subst. 
            inversion H9;subst. inversion H13; subst. inversion H15.
            unfold not in H2. apply H2. apply ex_intro with (x:=v). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H5. apply H5. apply ex_intro with (x:=v). auto.
          unfold not in n. apply n. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H13; subst.
            unfold not in n0. apply n0. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H6. apply H6. apply ex_intro with (x:=a). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H7. apply H7. apply ex_intro with (x:=e1). auto.
     (** t1 isnt v, t1 isnt av, t1 does not ex, t2 is v, t2 isnt av, t2 does ex *)
     assert ({t1=T_PF}+{~t1=T_PF}).
       apply eq_tm_dec.
     destruct H5. left. subst. apply value_T_PF.

      assert (decidable (exists u, t1 = tm_app P_SAYS u)).
       apply is_tm_app_dec2. 
      destruct H5.
       destruct H5 as [u H5]. subst.
       left. subst. apply value_P_SAYS; auto.

     assert ({t1=TM_PF_RETURN}+{~t1=TM_PF_RETURN}).
       apply eq_tm_dec.
      destruct H6. left. subst. apply value_TM_PF_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_RETURN u)).
       apply is_tm_app_dec2. 
      destruct H6.
       destruct H6 as [u H6]. subst.
       left. subst. apply value_PF_SAYS_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_BIND u)).
       apply is_tm_app_dec2. 
      destruct H7.
       destruct H7 as [u H7]. subst.
       left. subst. apply value_PF_SAYS_BIND; auto.

       right; unfold not; intro; inversion H8; subst.
          inversion H9; contradiction.
          inversion H8; subst. 
            inversion H9;subst. inversion H13; subst. inversion H15.
            unfold not in H2. apply H2. apply ex_intro with (x:=v). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H5. apply H5. apply ex_intro with (x:=v). auto.
          unfold not in n. apply n. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H13; subst.
            unfold not in n0. apply n0. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H6. apply H6. apply ex_intro with (x:=a). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H7. apply H7. apply ex_intro with (x:=e1). auto.
     (** t1 isnt v, t1 isnt av, t1 does not ex, t2 is v, t2 isnt av, t2 doesnt ex *)
     assert ({t1=T_PF}+{~t1=T_PF}).
       apply eq_tm_dec.
     destruct H5. left. subst. apply value_T_PF.

      assert (decidable (exists u, t1 = tm_app P_SAYS u)).
       apply is_tm_app_dec2. 
      destruct H5.
       destruct H5 as [u H5]. subst.
       left. subst. apply value_P_SAYS; auto.

     assert ({t1=TM_PF_RETURN}+{~t1=TM_PF_RETURN}).
       apply eq_tm_dec.
      destruct H6. left. subst. apply value_TM_PF_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_RETURN u)).
       apply is_tm_app_dec2. 
      destruct H6.
       destruct H6 as [u H6]. subst.
       left. subst. apply value_PF_SAYS_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_BIND u)).
       apply is_tm_app_dec2. 
      destruct H7.
       destruct H7 as [u H7]. subst.
       left. subst. apply value_PF_SAYS_BIND; auto.

       right; unfold not; intro; inversion H8; subst.
          inversion H9; contradiction.
          inversion H8; subst. 
            inversion H9;subst. inversion H13; subst. inversion H15.
            unfold not in H2. apply H2. apply ex_intro with (x:=v). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H5. apply H5. apply ex_intro with (x:=v). auto.
          unfold not in n. apply n. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H13; subst.
            unfold not in n0. apply n0. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H6. apply H6. apply ex_intro with (x:=a). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H7. apply H7. apply ex_intro with (x:=e1). auto.
     (** t1 isnt v, t1 isnt av, t1 does not ex, t2 isnt v, t2 is av, t2 does ex *)
     assert ({t1=T_PF}+{~t1=T_PF}).
       apply eq_tm_dec.
     destruct H5. left. subst. apply value_T_PF.

      assert (decidable (exists u, t1 = tm_app P_SAYS u)).
       apply is_tm_app_dec2. 
      destruct H5.
       destruct H5 as [u H5]. subst.
       left. subst. apply value_P_SAYS; auto.

     assert ({t1=TM_PF_RETURN}+{~t1=TM_PF_RETURN}).
       apply eq_tm_dec.
      destruct H6. left. subst. apply value_TM_PF_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_RETURN u)).
       apply is_tm_app_dec2. 
      destruct H6.
       destruct H6 as [u H6]. subst.
       left. subst. apply value_PF_SAYS_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_BIND u)).
       apply is_tm_app_dec2. 
      destruct H7.
       destruct H7 as [u H7]. subst.
       left. subst. apply value_PF_SAYS_BIND; auto.

       right; unfold not; intro; inversion H8; subst.
          inversion H9; contradiction.
          inversion H8; subst. 
            inversion H9;subst. inversion H13; subst. inversion H15.
            unfold not in H2. apply H2. apply ex_intro with (x:=v). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H5. apply H5. apply ex_intro with (x:=v). auto.
          unfold not in n. apply n. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H13; subst.
            unfold not in n0. apply n0. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H6. apply H6. apply ex_intro with (x:=a). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H7. apply H7. apply ex_intro with (x:=e1). auto.
     (** t1 isnt v, t1 isnt av, t1 does not ex, t2 isnt v, t2 is av, t2 doesnt ex *)
     assert ({t1=T_PF}+{~t1=T_PF}).
       apply eq_tm_dec.
     destruct H5. left. subst. apply value_T_PF.

      assert (decidable (exists u, t1 = tm_app P_SAYS u)).
       apply is_tm_app_dec2. 
      destruct H5.
       destruct H5 as [u H5]. subst.
       left. subst. apply value_P_SAYS; auto.

     assert ({t1=TM_PF_RETURN}+{~t1=TM_PF_RETURN}).
       apply eq_tm_dec.
      destruct H6. left. subst. apply value_TM_PF_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_RETURN u)).
       apply is_tm_app_dec2. 
      destruct H6.
       destruct H6 as [u H6]. subst.
       left. subst. apply value_PF_SAYS_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_BIND u)).
       apply is_tm_app_dec2. 
      destruct H7.
       destruct H7 as [u H7]. subst.
       left. subst. apply value_PF_SAYS_BIND; auto.

       right; unfold not; intro; inversion H8; subst.
          inversion H9; contradiction.
          inversion H8; subst. 
            inversion H9;subst. inversion H13; subst. inversion H15.
            unfold not in H2. apply H2. apply ex_intro with (x:=v). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H5. apply H5. apply ex_intro with (x:=v). auto.
          unfold not in n. apply n. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H13; subst.
            unfold not in n0. apply n0. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H6. apply H6. apply ex_intro with (x:=a). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H7. apply H7. apply ex_intro with (x:=e1). auto.
     (** t1 isnt v, t1 isnt av, t1 does not ex, t2 isnt v, t2 isnt av, t2 does ex *)
     assert ({t1=T_PF}+{~t1=T_PF}).
       apply eq_tm_dec.
     destruct H5. left. subst. apply value_T_PF.

      assert (decidable (exists u, t1 = tm_app P_SAYS u)).
       apply is_tm_app_dec2. 
      destruct H5.
       destruct H5 as [u H5]. subst.
       left. subst. apply value_P_SAYS; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_RETURN u)).
       apply is_tm_app_dec2. 
      destruct H6.
       destruct H6 as [u H6]. subst.
       left. subst. apply value_PF_SAYS_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_BIND u)).
       apply is_tm_app_dec2. 
      destruct H7.
       destruct H7 as [u H7]. subst.
       left. subst. apply value_PF_SAYS_BIND; auto.

       right; unfold not; intro; inversion H8; subst.
          inversion H9; contradiction.
          inversion H8; subst. 
            inversion H9;subst. inversion H13; subst. inversion H15.
            unfold not in H2. apply H2. apply ex_intro with (x:=v). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H5. apply H5. apply ex_intro with (x:=v). auto.
          unfold not in n. apply n. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H13; subst.
            unfold not in H0. apply H0 in H11. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H6. apply H6. apply ex_intro with (x:=a). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H7. apply H7. apply ex_intro with (x:=e1). auto.
     (** t1 isnt v, t1 isnt av, t1 does not ex, t2 isnt v, t2 isnt av, t2 doesnt ex *)
     assert ({t1=T_PF}+{~t1=T_PF}).
       apply eq_tm_dec.
     destruct H5. left. subst. apply value_T_PF.

      assert (decidable (exists u, t1 = tm_app P_SAYS u)).
       apply is_tm_app_dec2. 
      destruct H5.
       destruct H5 as [u H5]. subst.
       left. subst. apply value_P_SAYS; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_RETURN u)).
       apply is_tm_app_dec2. 
      destruct H6.
       destruct H6 as [u H6]. subst.
       left. subst. apply value_PF_SAYS_RETURN; auto.

      assert (decidable (exists u, t1 = tm_app PF_SAYS_BIND u)).
       apply is_tm_app_dec2. 
      destruct H7.
       destruct H7 as [u H7]. subst.
       left. subst. apply value_PF_SAYS_BIND; auto.

       right; unfold not; intro; inversion H8; subst.
          inversion H9; contradiction.
          inversion H8; subst. 
            inversion H9;subst. inversion H13; subst. inversion H15.
            unfold not in H2. apply H2. apply ex_intro with (x:=v). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H5. apply H5. apply ex_intro with (x:=v). auto.
          unfold not in n. apply n. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H13; subst.
            unfold not in H0. apply H0 in H11. auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H6. apply H6. apply ex_intro with (x:=a). auto.
          inversion H8;subst. 
            inversion H9;subst. inversion H12; subst. inversion H14.
            unfold not in H7. apply H7. apply ex_intro with (x:=e1). auto.
  split.
    Case "app av".
      decompose [and] IHt1. decompose [and] IHt2. clear IHt1 IHt2.
        destruct H; destruct H0; destruct H1; destruct H4;
        try solve[
           left;  apply applied_value_app; auto
          |right; unfold not; intro; inversion H4; contradiction
        ].
    Case "app ex".
      decompose [and] IHt1. decompose [and] IHt2. clear IHt1 IHt2.
      destruct H0.
      (* t2 is v*)
      assert ({t1=PF_SIGN}+{~t1=PF_SIGN}).
       apply eq_tm_dec.
      destruct H3. 
       left. subst. apply ex_intro with (x:=t2). auto.
       right. unfold not. intros.  destruct H3. destruct H3. inversion H3. subst.
         unfold not in n. apply n. auto.       
      (* t2 isnt v*)
      right. unfold not. intros.  destruct H3. destruct H3. inversion H3. subst.
         unfold not in H0. apply H0 in H6. auto.       
    Case "abs v".
    left. apply value_abs.
    Case "arr v".
    left. apply value_arr.
    Case "mat v".
    right. unfold not. intro. inversion H. inversion H0.
    Case "brn v".
    right. unfold not. intro. inversion H. inversion H0.
    Case "case v".
    right. unfold not. intro. inversion H. inversion H0.
Qed.

Lemma applied_value_dec: forall t, decidable (applied_value t).
Proof.
  intros. 
  decompose [and] (@value__applied_value__appPF_SIGN_dec t).
  auto.
Qed.

Lemma value_dec: forall t, decidable (value t).
Proof. 
  intros. 
  decompose [and] (@value__applied_value__appPF_SIGN_dec t).
  auto.
Qed.

(*********************************************************)
(** lc dec *)
Lemma swap_lc : forall x z t, 
  x \notin fv t -> z \notin fv t -> 
  lc (t^^tm_fv x) -> lc (t^^tm_fv z).
intros.
  rewrite (@subst_intro x); auto.
  eapply subst_lc; auto.
Qed.

Definition P_lc_dec (n:nat) := forall t, tm_size t = n ->decidable (lc t).
Lemma _lc_dec:  forall n, P_lc_dec n.
Proof.
  intro n.
  apply lt_wf_ind. clear n.
  intros n H.
  unfold P_lc_dec in *.
  unfold decidable in *.
  destruct t; intros; subst; try solve [right; unfold not; intro; inversion H0].
  left. apply lc_fv.
  left. apply lc_con.
  left. apply lc_ctr.

  assert (tm_size t1 < tm_size (tm_app t1 t2)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_app t1 t2)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H1; auto.
  destruct H0; destruct H1;
    try solve [right; unfold not; intro; inversion H2;contradiction].
    left. apply lc_app; auto.

  assert (tm_size t1 < tm_size (tm_abs t1 t2)). simpl. omega.
  pick_fresh y.
  assert (tm_size (t2^^(tm_fv y)) < tm_size (tm_abs t1 t2)).
    simpl. unfold open_tm. rewrite open_tm_size_eq. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2^^(tm_fv y)) in H1; auto.
  destruct H0; destruct H1.
    left. apply lc_abs with (L:=fv t1 \u fv t2); auto.
    intros. apply swap_lc with (x:=y); auto.

    right. unfold not. intro. inversion H2.
    pick_fresh x. 
    assert (x \notin L). auto. 
    apply H6 in H7.
    apply H1. apply swap_lc with (x:=x); auto.

    right. unfold not. intro. inversion H2. contradiction.
    right. unfold not. intro. inversion H2. contradiction.

  assert (tm_size t1 < tm_size (tm_arr t1 t2)). simpl. omega.
  pick_fresh y.
  assert (tm_size (t2^^(tm_fv y)) < tm_size (tm_arr t1 t2)).
    simpl. unfold open_tm. rewrite open_tm_size_eq. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2^^(tm_fv y)) in H1; auto.
  destruct H0; destruct H1.
    left. apply lc_arr with (L:=fv t1 \u fv t2); auto.
    intros. apply swap_lc with (x:=y); auto.

    right. unfold not. intro. inversion H2.
    pick_fresh x. 
    assert (x \notin L). auto. 
    apply H6 in H7.
    apply H1. apply swap_lc with (x:=x); auto.

    right. unfold not. intro. inversion H2. contradiction.
    right. unfold not. intro. inversion H2. contradiction.

  assert (tm_size t1 < tm_size (tm_mat t1 t2 t3)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_mat t1 t2 t3)). simpl. omega.
  assert (tm_size t3 < tm_size (tm_mat t1 t2 t3)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H1; auto.
  apply H with (t:=t3) in H2; auto.
  destruct H0; destruct H1; destruct H2;
    try solve [right; unfold not; intro; inversion H3;contradiction].
    left. apply lc_mat; auto.

  assert (tm_size t1 < tm_size (tm_brn c t1 t2)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_brn c t1 t2)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H1; auto.
  destruct H0; destruct H1;
    try solve [right; unfold not; intro; inversion H2;contradiction].
    left. apply lc_brn; auto.

  assert (tm_size t1 < tm_size (tm_cast t1 t2)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_cast t1 t2)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H1; auto.
  destruct H0; destruct H1;
    try solve [right; unfold not; intro; inversion H2;contradiction].
    left. apply lc_cast; auto.
Qed.

Lemma lc_dec:  forall t, decidable (lc t).
intros.
  assert (P_lc_dec (tm_size t)).
    apply (@_lc_dec (tm_size t)).
  unfold P_lc_dec in H.
  eapply H; eauto.
Qed.

(***********************************************************)
Lemma is_tm_type_dec: forall t,
  decidable (t = K_TYPE \/ t=K_PROP \/ t = S_KIND ).
Proof. 
  intros. unfold decidable.
  destruct t;
    try solve [
      right; unfold not; intro; 
        decompose [or] H;
        try solve [inversion H0|inversion H1]
    ].
    Case "con".
    destruct (01===c).
      (*S_KIND := tm_con 01*)
      left. subst. auto.
    destruct (10===c).
      (*K_TYPE := tm_con 10*)
      left. subst. auto.
    destruct (11===c).
      (*K_PROP := tm_con 11*)
      left. subst. auto.
    right; unfold not; intro. 
    decompose [or] H. 
      inversion H0. subst. unfold not in n0. auto.
      inversion H1. subst. unfold not in n1. auto.
      inversion H1. subst. unfold not in n. auto.
Qed.

Lemma fully_applied_args_dec: forall t,
 decidable (exists ctr, exists args, fully_applied_args t ctr args).
Proof.
 intros. unfold decidable.
 induction t;
   try solve [right; unfold not; intros; destruct H; destruct H; inversion H].
   left. apply ex_intro with (x:=c). apply ex_intro with (x:=(@nil tm)). auto.
   destruct IHt1; destruct IHt2.
     destruct H as [ctr1 H]. 
     destruct H as [args1 H]. 
     left. apply ex_intro with (x:=ctr1). apply ex_intro with (x:=(t2::args1)). auto.

     destruct H as [ctr1 H]. 
     destruct H as [args1 H]. 
     left. apply ex_intro with (x:=ctr1). apply ex_intro with (x:=(t2::args1)). auto.

     right. unfold not. intros.
     destruct H1 as [ctr12 H1]. 
     destruct H1 as [args12 H1].
     inversion H1. subst.
     unfold not in H. apply H.
     apply ex_intro with (x:=ctr12). apply ex_intro with (x:=args). auto. 
 
     right. unfold not. intros.
     destruct H1 as [ctr12 H1]. 
     destruct H1 as [args12 H1].
     inversion H1. subst.
     unfold not in H. apply H.
     apply ex_intro with (x:=ctr12). apply ex_intro with (x:=args). auto. 
Qed.

Lemma fully_applied_args_dec2: forall t ctr,
 decidable (exists args, fully_applied_args t ctr args).
Proof.
 intros. unfold decidable. gen ctr.
 induction t; intro;
   try solve [right; unfold not; intros; destruct H; inversion H].
   destruct (eq_tm_dec (tm_ctr c) (tm_ctr ctr)).
     inversion e. subst. left. apply ex_intro with (x:=(@nil tm)). apply fully_applied__args_ctr.
     right. unfold not. intros. inversion H. inversion H0. subst. apply n; auto.
   destruct (@IHt1 ctr); destruct (@IHt2 ctr).
     destruct H as [args1 H]. 
     left. apply ex_intro with (x:=(t2::args1)). auto.

     destruct H as [args1 H]. 
     left. apply ex_intro with (x:=(t2::args1)). auto.

     right. unfold not. intros.
     destruct H1 as [args12 H1].
     inversion H1. subst.
     unfold not in H. apply H.
     apply ex_intro with (x:=args). auto. 
 
     right. unfold not. intros.
     destruct H1 as [args12 H1].
     inversion H1. subst.
     unfold not in H. apply H.
     apply ex_intro with (x:=args). auto. 
Qed.

Lemma fully_applied_typ_dec: forall t ctr k,
  decidable (fully_applied_typ t ctr k).
Proof.
 intros. unfold decidable.
 gen ctr k.
 induction t; intros;
   try solve [right; unfold not; intros; inversion H].

   assert ({k=K_TYPE}+{~k=K_TYPE}).
       apply eq_tm_dec.
   destruct c. destruct ctr.
   destruct H; destruct (n===n1); destruct (n0===n2);
      try solve [subst; right; unfold not; intro; inversion H; subst; unfold not in n3; auto].
      subst. left. apply fully_applied_TYPE. 
   assert ({k=K_PROP}+{~k=K_PROP}).
       apply eq_tm_dec.
   destruct H.
      subst. left. apply fully_applied_PROP. 
      subst. right. unfold not. intros. inversion H.
        subst. unfold not in n3. auto. 
        subst. unfold not in n4. auto. 

  destruct k;
    try solve [right; unfold not; intros; inversion H].
    destruct (@IHt1 ctr k2).
      left. apply fully_applied_typ_app; auto.
      right. unfold not. intros. inversion H0; subst. apply H; auto.
Qed.

Lemma is_tm_arr_dec': forall t, 
   decidable (exists t1, exists t2, t = tm_arr t1 t2).
Proof. 
  intros. unfold decidable. 
  destruct t;
    try solve [right; unfold not; intro; destruct H as [u1 H]; destruct H as [u2 H]; inversion H].
    left. apply ex_intro with (x:=t1). apply ex_intro with (x:=t2). reflexivity.
Qed.

Lemma wf_brn_rename: forall S1 S2 s t u r x y,
  wf_sig S1 ->
  wf_brn S1 S2 s nil (t^^tm_fv x) (u^^tm_fv x) r ->
  x\notin fv s \u fv t \u fv u \u fv r ->
  y\notin fv s \u fv t \u fv u \u fv r ->
  wf_brn S1 S2 s nil (t^^tm_fv y) (u^^tm_fv y) r.
Proof. 
 introv Hwf_sig Hwfx Hx Hy.
 rewrite (@subst_intro x); auto.
  rewrite (@subst_intro x u); auto.
  rewrite <- (@subst_fresh x (tm_fv y) s); auto.
  rewrite <- (@subst_fresh x (tm_fv y) r); auto.
  replace (@nil tm) with (List.map (subst_tm x (tm_fv y)) nil); auto.
  eapply (proj1 (proj2 (proj2 (substitution S2 Hwf_sig )))); eauto.
Qed.
	
Lemma wf_brn_dec: forall S1 S2 s args t u r,
  wf_sig S1 -> lc t  -> all_tm_lc args -> decidable ( wf_brn S1 S2 s args t u r).
Proof.
introv Hwf_sig Hlct Hlcargs.
gen t u.
induction Hlcargs; intros.
Case "args = nil". 
 gen u. induction Hlct; intros. 
 Case "fv".
 destruct (eq_tm_dec s (tm_fv a)).
 destruct (eq_tm_dec u r).
 left; subst; auto.
Case "u<>r". 
    right; intro J; apply n; inversion J; subst; auto.
Case "s<>tm_fv a".
    right; intro J; apply n; inversion J; subst; auto.
Case   "cons". 
   destruct (eq_tm_dec s (tm_con a)).
   destruct (eq_tm_dec u r).
   left; subst; auto.
   Case "u<>r". 
   right; intro J; apply n; inversion J; subst; auto.
   Case "s<>tm_fv a".
   right; intro J; apply n; inversion J; subst; auto.
Case "ctr". 
   destruct (eq_tm_dec s (tm_ctr a)).
   destruct (eq_tm_dec u r).
   left; subst; auto.
   Case "u<>r". 
   right; intro J; apply n; inversion J; subst; auto.
   Case "s<>tm_fv a".
   right; intro J; apply n; inversion J; subst; auto.
Case "app". 
   destruct (eq_tm_dec s (tm_app t1 t2)).
   destruct (eq_tm_dec u r).
   left; subst; auto.
   Case "u<>r". 
   right; intro J; apply n; inversion J; subst; auto.
   Case "s<>tm_fv a".
   right; intro J; apply n; inversion J; subst; auto.
Case "abs".
     destruct (eq_tm_dec s (tm_abs t1 t2)).
   destruct (eq_tm_dec u r).
   left; subst; auto.
   Case "u<>r". 
   right; intro J; apply n; inversion J; subst; auto.
   Case "s<>tm_fv a".
   right; intro J; apply n; inversion J; subst; auto.
Case "arr".
  destruct (eq_tm_dec s (tm_arr t1 t2)).
  destruct (eq_tm_dec u r).
  left; rewrite e. rewrite e0.  auto.
  Case "u<>r". 
  destruct (is_tm_arr_dec u t1). destruct H1 as [u0 Hu].
   subst.
   pick_fresh z.
  assert (J: decidable (wf_brn S1 S2 (tm_arr t1 t2) nil (t2^^tm_fv z) (u0^^tm_fv z) r)).
  eapply H0; eauto.
  destruct J.
  gather_atoms' as L'.
  subst; left; eapply wf_brn_constructor_arg with (L:=L'); eauto.
  intros. unfold L' in H2.
  eapply wf_brn_rename; eauto.  simpl; auto. simpl; auto.  
  right; intro J; inversion J;  subst; eauto.   apply H1. 
  pick_fresh y. eapply wf_brn_rename with (x:=y); eauto. simpl; auto. simpl; auto.

  Case "u<> t1 -> u0".
  right; intro J; inversion J; subst;  eauto.
  Case "s<>t1 -> t2".
  destruct (is_tm_arr_dec u t1). destruct H1 as [u0 Hu].
   subst.
   pick_fresh z.
   assert (J: decidable (wf_brn S1 S2 s nil (t2^^tm_fv z) (u0^^tm_fv z) r)).
   eapply H0; eauto.
   destruct J.
   gather_atoms' as L'.
   subst; left; eapply wf_brn_constructor_arg with (L:=L'); eauto.
   intros. unfold L' in H2.  eapply wf_brn_rename with (x:=z); eauto.
   right; intro J; inversion J;  subst; eauto.   apply H1. 
     pick_fresh y. eapply wf_brn_rename with (x:=y); eauto.
   Case "u<> t1 -> u0".
   right; intro J; inversion J; subst;  eauto.

 Case "mat".
   destruct (eq_tm_dec s (tm_mat t1 t2 t3)).
   destruct (eq_tm_dec u r).
   left; subst; auto.
   Case "u<>r". 
   right; intro J; apply n; inversion J; subst; auto.
   Case "s<>tm_fv a".
   right; intro J; apply n; inversion J; subst; auto.
 Case "brn". 
     destruct (eq_tm_dec s (tm_brn c t2 t3)).
   destruct (eq_tm_dec u r).
   left; subst; auto.
   Case "u<>r". 
   right; intro J; apply n; inversion J; subst; auto.
   Case "s<>tm_fv a".
   right; intro J; apply n; inversion J; subst; auto.
 Case "cast". 
      destruct (eq_tm_dec s (tm_cast t1 t2)).
   destruct (eq_tm_dec u r).
   left; subst; auto.
   Case "u<>r". 
   right; intro J; apply n; inversion J; subst; auto.
   Case "s<>tm_fv a".
   right; intro J; apply n; inversion J; subst; auto.

Case "args = a::args'".
  assert (J:= is_tm_arr_dec' t). 
  destruct J. destruct H0 as [t1 [t2 Jt]].
  subst. assert (J: decidable (wf_brn S1 S2 s l (t2^^a) u r)).
  eapply IHHlcargs; eauto. 
  inversion Hlct; subst.
  pick_fresh z. assert (J: lc (t2^^tm_fv z)); eauto.
  rewrite (@subst_intro z); auto.  apply subst_lc; auto. 
  destruct J.
  left; eauto.
  Case "~wf_brn S s args (t2^^a) u r".
  right; intro J; inversion J; subst; apply H0; eauto.
  Case "~t = t1 -> t2".
   right; intro J; inversion J; subst; apply H0; eauto.
Qed.

(******************************************************)
(** branches_cover_dec *)
Lemma _branches_cover_aux_dec: forall c b,
 decidable (branches_cover_aux c b).
Proof.
 introv. gen b. unfold decidable.
 induction c; intro. 
   destruct b; try solve [right; unfold not; intros; inversion H].
   destruct (c === 00).
     left. subst. apply bca_nil.
     right; unfold not; intros; inversion H. subst. apply n; auto.

   destruct b; try solve [right; unfold not; intros; inversion H].
   destruct (eq_tm_dec (tm_brn a b1 b2) (tm_brn c0 b1 b2)).
     inversion e. subst. destruct (@IHc b2).
       left. apply bca_cons; auto.
       right. unfold not. intros. inversion H0. apply H; auto.
     right. unfold not. intros. inversion H; subst. 
       unfold not in n. apply n; auto.
Qed.

Lemma branches_cover_aux_dec: forall b,
 decidable (exists c, branches_cover_aux c b).
Proof.
 introv. unfold decidable.
 induction b;
   try solve [right; unfold not; intros; inversion H; inversion H0].
   destruct (c === 00).
     left. subst. apply ex_intro with (x:=(@nil ctr)). apply bca_nil.
     right; unfold not; intros; inversion H. inversion H0. subst. apply n; auto.

   destruct IHb2. 
     destruct H; subst.
       left. apply ex_intro with (x:=c::x). apply bca_cons; auto.
       right. unfold not. intros. inversion H0. inversion H1; subst.
         apply H. apply ex_intro with (x:=l). auto.
Qed.

Lemma get_defns_ctrs_of_some_dec: forall c d,
  decidable (exists l, get_defns_ctrs_of c d = Some l).
intros. gen c.
  induction d; intros; simpl; unfold decidable.
    right. intro. inversion H. inversion H0.

    destruct a. destruct p.
      destruct (eq_ctr_dec c c1); subst.
        destruct c0.
          simpl. left. apply ex_intro with (x:=(@nil ctr)). reflexivity.
          simpl. destruct p.
            left. apply ex_intro with (x:=c::dom_ctr_decls c0). reflexivity.
        destruct (@IHd c).
          left. auto.
          right. auto. 
Qed.

Lemma get_defns_ctrs_of_none_dec: forall c d,
  (~ exists l, get_defns_ctrs_of c d = Some l) ->
  decidable (get_defns_ctrs_of c d = None).
intros. gen c.
  induction d; intros; simpl; unfold decidable.
    left. auto.

    destruct a. destruct p.
      destruct (eq_ctr_dec c c1); subst.
        right. unfold not. intros. inversion H0.

        simpl in H.
        destruct (eq_ctr_dec c c1); subst.
          unfold not in n. assert (c1=c1). auto. apply n in H0. inversion H0.

          apply IHd in H. auto.
Qed.

Lemma get_defns_ctrs_of_false: forall c d,
  (~ exists l, get_defns_ctrs_of c d = Some l)  /\ (get_defns_ctrs_of c d <> None) ->
  False.
intros.
  gen c.
  induction d; intros; destruct H.
    assert (get_defns_ctrs_of c nil = None). auto. apply H0; auto.

   simpl in H0. simpl in H. destruct a. destruct p.
   destruct (eq_ctr_dec c c1).
     apply H.  apply ex_intro with (x:=(dom_ctr_decls c0)); auto.
     apply (@IHd c); auto.
Qed.

Lemma branches_cover_dec: forall S b ctr,
  decidable (branches_cover S b ctr).
Proof.
introv. unfold decidable. unfold branches_cover.
 gen b ctr. induction S; intros.
   simpl. right. unfold not. intros. inversion H. inversion H0. inversion H1. 

   simpl. destruct a.  
     destruct (get_defns_ctrs_of_some_dec ctr d).
       inversion H; subst. rewrite H0.
       destruct (_branches_cover_aux_dec x b).
           left. apply ex_intro with (x:=x). auto.
           right. unfold not. intros. inversion H2. inversion H3. inversion H4; subst.
             apply H1; auto.
       
       assert (~exists l, get_defns_ctrs_of ctr d = Some l). auto.     
       apply get_defns_ctrs_of_none_dec in H.
       destruct H. 
         rewrite H.
         destruct (@IHS b ctr).
           left. auto.
           right. auto.

         assert (False). apply get_defns_ctrs_of_false with (c:=ctr) (d:=d); auto.
         inversion H1.

     destruct (@IHS b ctr).
       left. auto.
       right. auto.
Qed.

(******************************************************)
Inductive atomic : tm -> Prop:=
 | atomic_fv : forall x, atomic (tm_fv x)
 | atomic_con: forall c, atomic (tm_con c)
 | atomic_ctr: forall c, atomic (tm_ctr c).

Hint Constructors atomic. 

Lemma  ctr_has_arr_typ: forall  S1 S2 E t1 t2 k c args,
 wf_tm S1 S2 E (tm_app t1 t2) k -> fully_applied_args_ctr t1 c args ->
 exists s1, exists s2, wf_tm S1 S2 E (tm_ctr c) (tm_arr s1 s2).
Proof.
 intros.
 gen t2 k.
 induction H0; intros; try solve [inversion H; subst; eauto].
 inversion H; subst.
 eapply IHfully_applied_args_ctr with (t2:= tn) ; eauto.
 inversion H0.
 inversion H0. inversion H0. inversion H0. inversion H11. inversion H18. 
 inversion H0. inversion H0.
Qed.
 
Lemma sig_lookup_later: forall c d S,
 ~In c (dom_defns d) -> sig_lookup_ctr c (bundle_defns d::S) = sig_lookup_ctr c S.
Proof.
 intros.
 simpl.
 remember (bundle_defns_ctr_lookup c d) as B.
 destruct B; auto.
 symmetry in HeqB.
 destruct H; eapply defns_lookup_some_in; eauto.
 Qed.

Lemma in_get_atomic_ctr_in_sig: forall S L L1 c,
  get_atomic_ctrs_sig S L = L1 -> In c L1 -> In c L \/ In c (dom_sig S).
Proof.
 intros. 
 gen L1.
  induction S; intros.
  inversion H; subst; auto.
  destruct a. 
   simpl in H.
    remember (all_atomic_decls_defns(dom_defns(get_tctr_defns_of d) ++get_atomic_ctrs_sig S L) d) as A.
    destruct A.
    rewrite <- H in H0.
    unfold dom_sig; simpl.
    assert (J: In c (get_atomic_ctrs_sig S L) \/ In c (dom_defns(get_tctr_defns_of d))).
    apply in_app_or; auto. replace (flat_map dom_bundle S) with (dom_sig S); auto.
    assert  (Jin :In c L \/In c (dom_defns d) \/ In c (dom_sig S)).
    destruct J; auto.
    destruct (IHS (get_atomic_ctrs_sig S L)); auto.
    right; left; eapply in_tctr_defns_in; eauto.
    destruct Jin; auto.
    right; eapply in_or_app; eauto.
    
    unfold dom_sig; simpl.
    assert  (Jin :In c L \/In c (dom_defns d) \/ In c (dom_sig S)).
    destruct (IHS L1); auto.
    destruct Jin; auto.
    right; eapply in_or_app; eauto.  
    
  simpl. simpl in H.
  destruct (IHS L1 H H0); auto.
Qed.

Lemma in_tctr_some: forall c d, 
 In c (dom_defns(get_tctr_defns_of d)) -> exists t, bundle_defns_ctr_lookup c d = Some t.
Proof.
intros.
induction d; auto.
simpl in H. destruct H.

destruct a. destruct p. 
simpl in H.
destruct H; subst; auto.
exists t;simpl; auto.
destruct (eq_ctr_dec c c); subst; auto.
destruct n; auto.

 simpl. destruct (eq_ctr_dec c c1); subst; eauto.
 destruct (ctr_decls_lookup c c0); subst; eauto.
Qed.

Lemma all_atomic_defns_TYPE: forall L c d t, 
  all_atomic_decls_defns L  d = true -> 
  In c (dom_defns (get_tctr_defns_of d)) -> 
  wf_dom_defns d -> 
  bundle_defns_ctr_lookup c d = Some t -> 
  t = K_TYPE.
Proof.
 intros.
 induction d.
 simpl in H0. destruct H0.

 destruct a. destruct p.
 inversion H1; subst.
 simpl in H0.  simpl in H2. simpl in H.
 destruct (eq_ctr_dec c c1); subst; auto.
 inversion H2; subst.
 destruct (eq_tm_dec t K_TYPE); subst; auto. inversion H.

 remember (ctr_decls_lookup c c0) as C.
 destruct C. symmetry in HeqC.
 assert (Jin: In c (dom_ctr_decls c0)). eapply ctr_decls_lookup_some_in; eauto.
 assert (Jin1: In c (dom_defns d)).
 destruct H0; auto. destruct n; auto.
 eapply in_tctr_defns_in; eauto.
 destruct (H8 c Jin Jin1).

 destruct H0. destruct n; auto.
 destruct (eq_tm_dec t0 K_TYPE). 
 inversion H.
 eapply IHd; eauto. 
 assert (J: (all_atomic_ctr_types L c0 = true) /\ (all_atomic_decls_defns L d = true)).
 apply Bool.andb_prop; auto.  destruct J; auto.
 inversion H.
Qed.

Lemma in_get_atomic_ctr_TYPE: forall S2 L L1 c, 
  get_atomic_ctrs_sig S2 L1 = L -> In c L -> wf_dom_sig S2 -> In c L1 \/ sig_lookup_ctr c S2 = Some K_TYPE.
Proof.
 intros.
 gen L1 L.
 induction S2; intros.
 inversion H; subst; auto.
 
 destruct a. 
  inversion H1; subst.
 simpl in H0.
 remember (all_atomic_decls_defns(dom_defns(get_tctr_defns_of d) ++get_atomic_ctrs_sig S2 L1) d) as A.
 destruct A.
 assert (J: In c (get_atomic_ctrs_sig S2 L1) \/ In c (dom_defns(get_tctr_defns_of d))). eapply in_app_or; eauto.
 destruct J; auto.
 assert (Jin: In c L1 \/ In c (dom_sig S2)). eapply  in_get_atomic_ctr_in_sig; eauto.
 destruct Jin ;auto.
  replace (sig_lookup_ctr c (bundle_defns d::S2)) with (sig_lookup_ctr c S2); auto.
 eapply IHS2; auto.
 symmetry. eapply sig_lookup_later; eauto.
 unfold not; intro J; destruct (H6 c J H2).
 
  right. simpl.
  remember (bundle_defns_ctr_lookup c d) as B.
  destruct B. symmetry in HeqB. 
  assert (Jeq: t = K_TYPE); subst; auto.
  eapply  all_atomic_defns_TYPE; eauto.

  assert (exists t, bundle_defns_ctr_lookup c d = Some t).
 eapply in_tctr_some; eauto. destruct H2. rewrite H2 in HeqB. inversion HeqB.

  assert (Jin: In c L1 \/ In c (dom_sig S2)). eapply  in_get_atomic_ctr_in_sig; eauto.
 destruct Jin ;auto.
  replace (sig_lookup_ctr c (bundle_defns d::S2)) with (sig_lookup_ctr c S2); auto.
 eapply IHS2; auto.
 symmetry. eapply sig_lookup_later; eauto.
 unfold not; intro J; destruct (H6 c J H).
 
 inversion H1; subst.
 simpl in H0. simpl; eauto. 
Qed.

Lemma wf_ctr_unique': forall S ctr t1 t2,
 wf_dom_sig S 
-> sig_has_ctr ctr t1 S
-> sig_has_ctr ctr t2 S
-> t1 = t2.
Proof.
intros. 
 induction H.
 Case "nil". inversion H0.
 Case "assn". 
      unfold sig_has_ctr in H0; simpl in H0.
      unfold sig_has_ctr in H1; simpl in H1.
      apply IHwf_dom_sig; auto.
 Case "defn_type". 
       unfold sig_has_ctr in H0; simpl in H0.
       remember (bundle_defns_ctr_lookup ctr d) as B.
       destruct B.
       Case "B = some". 
          unfold sig_has_ctr in H1; simpl in  H1.
            remember (bundle_defns_ctr_lookup ctr d) as B1.
            destruct B1. inversion H0.  inversion H1. inversion HeqB; subst; auto.
            inversion HeqB.
       Case "B = None". 
           unfold sig_has_ctr in H1; simpl in  H1.
            remember (bundle_defns_ctr_lookup ctr d) as B1.
            destruct B1.  inversion HeqB.
            apply IHwf_dom_sig; auto.
Qed.

(* TODO *)
Lemma converts_dec: forall S1 S2 E t1 t2,
  wf_env S1 S2 E ->  decidable (converts E t1 t2).
Proof.
 Admitted.


Lemma open_eq_const: forall u z c,
    u^^tm_fv z = tm_con c -> u = tm_con c.
Proof. 
  intros. destruct u;
    try solve [unfold open_tm in H; simpl in H; inversion H].

    unfold open_tm in H; simpl in H.
    destruct (b===0); inversion H.
     
    unfold open_tm in H; simpl in H.
    auto.
Qed.


Lemma decompose_open: forall z k n,
  (exists u, k = open_tm_aux n (tm_fv z) u /\ z \notin fv u) \/ (~exists u, k = open_tm_aux n (tm_fv z) u).
Proof. 
 introv.
 gen n. 
 induction k; intro.
  Case "fv".
  left.
  destruct (a ==z).
  subst. exists (tm_bv n); split.
   simpl. destruct (n===n); auto. assert (J: n=n); auto; contradiction.
   simpl; auto.
   exists (tm_fv a); split; simpl; auto.

 Case "bv".
  destruct (b===n).
  subst.  
  right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
  destruct (b===n); inversion Jeq. symmetry in H0; contradiction.

  left; exists (tm_bv b); split; simpl; auto.
  destruct (b===n); tauto.

Case "con".
 left; exists (tm_con c); split; simpl; auto; destruct (b===n); tauto.

 left; exists (tm_ctr c); split; simpl; auto; destruct (b===n); tauto.

Case "app".
 destruct (IHk1 n).  destruct (IHk2 n).
 destruct H as [u1 [J1 J2]].   destruct H0 as [u2 [J3 J4]].
 left; exists (tm_app u1 u2); subst; simpl; split; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H0; exists u2; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H; exists u1; auto.

Case "abs". 
 destruct (IHk1 n).  destruct (IHk2 (1+n)).
 destruct H as [u1 [J1 J2]].   destruct H0 as [u2 [J3 J4]].
 left; exists (tm_abs u1 u2); subst; simpl; split; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H0; exists u2; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H; exists u1; auto.


Case "arr". 
destruct (IHk1 n).  destruct (IHk2 (1+n)).
 destruct H as [u1 [J1 J2]].   destruct H0 as [u2 [J3 J4]].
 left; exists (tm_arr u1 u2); subst; simpl; split; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H0; exists u2; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H; exists u1; auto.

Case "tm_mat".  
 destruct (IHk1 n).  destruct (IHk2 n). destruct (IHk3 n).
 destruct H as [u1 [J1 J2]].   destruct H0 as [u2 [J3 J4]]. destruct H1 as [u3 [J5 J6]].
 left; exists (tm_mat u1 u2 u3); subst; simpl; split; auto.

  right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H1; exists u3; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H0; exists u2; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H; exists u1; auto.

Case "tm_brn".
 destruct (IHk1 n).  destruct (IHk2 n).
 destruct H as [u1 [J1 J2]].   destruct H0 as [u2 [J3 J4]].
 left; exists (tm_brn c u1 u2); subst; simpl; split; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H0; exists u2; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H; exists u1; auto.

Case "tm_cast".
 destruct (IHk1 n).  destruct (IHk2 n).
 destruct H as [u1 [J1 J2]].   destruct H0 as [u2 [J3 J4]].
 left; exists (tm_cast u1 u2); subst; simpl; split; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H0; exists u2; auto.

 right; intro J; destruct J as [u Jeq].
  destruct u; subst; simpl in Jeq; try solve [inversion Jeq].
    destruct (b===n); inversion Jeq.
    inversion Jeq; subst. clear IHk1 IHk2.
    apply H; exists u1; auto.
Qed.

Lemma subst_id: forall x u,   ([x~> tm_fv x]u) = u.
Proof.
 introv.
 induction u; simpl; auto; try solve [ rewrite IHu1; rewrite IHu2; auto].
 
 destruct (a ==x); subst; auto.
 rewrite IHu1; rewrite IHu2; rewrite IHu3; auto.
Qed.

 
Lemma converts_contraction_eq: forall S1 S2 E x y v1 v2 k F t1 t2, 
 converts (E&x~bnd_eq v1 v2 k & y~bnd_eq v1 v2 k &F) t1 t2 -> 
 wf_env S1 S2 (E&y~bnd_eq v1 v2 k &F) -> converts (E&y~bnd_eq v1 v2 k &F) t1 t2.
Proof. 
 intros.
 remember (E&x~bnd_eq v1 v2 k & y~bnd_eq v1 v2 k &F)  as E0.
 induction H.
 apply conv_refl; auto.
 apply conv_sym; auto.
 apply conv_trans with (t2:=t2); auto.
 Case "axiom".
   rewrite HeqE0 in H. binds_cases H.
   apply conv_axiom with (x:=x0) (k:=k0); auto.
   rewrite_env (E&(  y~bnd_eq v1 v2 k & F)&empty).  replace E with (E&empty) in B ; auto.
   eapply binds_eq_weakening; eauto.
   rewrite_env (E& y~bnd_eq v1 v2 k & F); eauto.
   
   inversion H1; subst. 

  eapply conv_axiom with (x:=y) (k:=k); eauto. 
  unfold binds_eq. 
  apply binds_mid; auto. eapply wf_env_ok; eauto.
   inversion H1; subst. eapply conv_axiom with (x:=x0) (k:=k); eauto.
   rewrite_env ((E&x0~bnd_eq v1 v2 k) & F & empty).
   eapply binds_eq_weakening; eauto.
   rewrite_env (empty& E& x0~bnd_eq v1 v2 k). 
   eapply binds_eq_weakening; eauto.
  rewrite_env (E& x0~bnd_eq v1 v2 k); eauto.  eapply wf_env_contraction; eauto.
  eapply conv_axiom with (x:=x0) (k:=k0); eauto.
   rewrite_env (empty&(E& y~bnd_eq v1 v2 k) &F).  replace F with (empty& F) in B0 ; auto.
   eapply binds_eq_weakening; eauto. 
   rewrite_env (E&y~bnd_eq v1 v2 k& F); eauto. apply empty_concat.

 apply conv_app; auto.
 Case "abs".
    gather_atoms' as L'.
    apply conv_abs with (L:=L'); auto.
    intros. unfold L' in H3. eapply H2; eauto.

Case "arr".
    gather_atoms' as L'.
    apply conv_arr with (L:=L'); auto.
    intros. unfold L' in H3. eapply H2; eauto.

Qed.

Lemma typing_contraction_eq_mutual:
forall S1  S2,
wf_sig S1 ->
((forall E0 e t,  wf_tm S1 S2 E0 e t ->
  forall E F x y v1 v2 k,  E0 = (E& x~ bnd_eq v1 v2 k & y~bnd_eq v1 v2 k & F) ->
   wf_env S1 S2 (E& y~ bnd_eq v1 v2 k & F) ->
   wf_tm S1 S2 (E & y~bnd_eq v1 v2 k & F) e t)
/\ (forall E0 s args branches r_tp,  wf_branches S1 S2 E0 s args branches r_tp ->
             forall E F x y v1 v2 k,  E0 = (E& x~ bnd_eq v1 v2  k&y~ bnd_eq v1 v2  k& F) -> 
             wf_env S1 S2 (E & y~ bnd_eq v1 v2 k &F) ->
            wf_branches S1 S2 (E&y~ bnd_eq v1 v2 k &F) s args branches r_tp )).
Proof.
intros.
 apply  mutual_wf_tm_branches_ind; intros; auto. 

  Case "arr".
  gather_atoms' as L'.
  eapply wf_tm_arr with (L:=L'); eauto.
  intros. unfold L' in H4. 
  rewrite_env (E0 & y~ bnd_eq v1 v2 k& (F& x0~:u1)). eapply H1 with (x0:=x) (x:=x0) (k:=k); eauto.  rewrite H2; auto.
  rewrite_env ((E0 & y~bnd_eq v1 v2 k & F) & (x0~:u1)).
  eapply wf_env_cons_var; eauto.

 Case "fv".
  rewrite H0 in b. binds_cases b. 
 apply wf_tm_var; auto.
  replace E0 with (E0 & empty) in B; auto. rewrite_env ( E0 & (y~bnd_eq v1 v2 k& F) & empty).
  eapply binds_var_weakening; eauto.
  rewrite_env ( E0 & y~bnd_eq v1 v2 k& F); eauto.
  apply wf_tm_var; auto.
  rewrite_env (empty&(E0 & y~ bnd_eq v1 v2 k)&F).
  replace F with (empty&F) in B0; auto. eapply binds_var_weakening; eauto.
  rewrite_env (E0&y~bnd_eq v1 v2 k & F); eauto. apply empty_concat.

Case "abs".
  gather_atoms' as L'.
 eapply wf_tm_abs with (L:=L'); eauto.
 intros. unfold L' in H5. rewrite_env (E0 & y~bnd_eq v1 v2 k0 & (F&x0 ~:u1)). 
  eapply H1 with (x0:=x) (F:= F&x0~:u1) (k:=k0)  ; eauto. 
  rewrite H3; eauto. 
  rewrite_env (E0 &y~bnd_eq v1 v2 k0 & F&x0~:u1); eapply wf_env_cons_var; eauto.

Case "app".
  eapply wf_tm_app with (t2:=t2); eauto.
Case "SAYS".
  eapply wf_tm_P_SAYS; eauto.
   eapply wf_tm_T_PF; eauto.
   eapply wf_tm_TM_PF_RETURN; eauto.
   eapply wf_tm_TM_PF_BIND; eauto.
   eapply wf_tm_TM_SAY; eauto.

Case "pif".
   gather_atoms' as L'.
   eapply wf_tm_TM_PIF with (L:=L'); eauto.
   intros. unfold L' in H7. rewrite_env (E0 & y~ bnd_eq v3 v4 k & (F& x0~bnd_eq v1 v2 T1)). 
   eapply H2 with (F:=F &x0~bnd_eq v1 v2 T1) (x:=x0)  (k:=k) (x0:=x); eauto.
   rewrite H5; auto.
   rewrite_env (E0 &y~bnd_eq v3 v4 k&  F &x0~bnd_eq v1 v2 T1).
   apply wf_env_cons_eq; auto.
  
  eapply H0;eauto.
  eapply H1; eauto. eapply H4; eauto.

 eapply wf_tm_PF_SAYS_RETURN; eauto.
  eapply wf_tm_PF_SAYS_BIND; eauto.

 eapply wf_tm_mat_TYPE; eauto.
 eapply wf_tm_mat_PROP; eauto.
  eapply wf_tm_cast; eauto. rewrite H2 in c. 
  eapply converts_contraction_eq; eauto.
  eapply wf_branches_Some; eauto.
Qed.
  
 
Lemma typing_eq_rename: forall S1 S2 E x y  v1 v2 k t u,
 wf_sig S1 ->
 wf_tm S1 S2 (E&x~bnd_eq v1 v2 k) t u->
 y \notin fv t  \u dom E ->
 x \notin fv t  \u dom E  ->
 wf_tm S1 S2 (E&y~bnd_eq v1 v2 k) t u.
Proof. 
introv Hwf_sig Hwfx Hx Hy.
destruct (x==y).
subst; auto. 

assert (Jwf_envx:= wf_tm_gives_wf_env Hwfx).
inversion Jwf_envx; subst.
assert (Jwf_env: wf_env S1 S2 (E& x ~ bnd_eq v1 v2 k & y~bnd_eq v1 v2 k)).
eapply wf_env_cons_eq; eauto.
eapply typing_weakening'; eauto.
eapply typing_weakening'; eauto.
eapply typing_weakening'; eauto.
assert (J: wf_tm S1 S2 (E& x ~ bnd_eq v1 v2 k & y~bnd_eq v1 v2 k) t u).
eapply typing_weakening'; eauto.
rewrite_env (E& y~bnd_eq v1 v2 k& empty).
replace (E& x~bnd_eq v1 v2 k& y~bnd_eq v1 v2 k) with  (E& x~bnd_eq v1 v2 k& y~bnd_eq v1 v2 k& empty ) in J; auto.
eapply (proj1 (typing_contraction_eq_mutual S2 Hwf_sig)) with (x:=x); eauto.
rewrite_env (E& y~bnd_eq v1 v2 k); auto.
Qed.


Lemma typing_rename: forall S1 S2 E x y k t u,
 wf_sig S1 ->
 wf_tm S1 S2 (E&x~:k) (t^^tm_fv x) u->
 y \notin fv t  \u dom E  ->
 x \notin fv t  \u dom E  ->
 wf_tm S1 S2 (E&y~:k) (t^^tm_fv y) ([x~> tm_fv y]u).
Proof. 
introv Hwf_sig Hwfx Hx Hy.
destruct (x==y).
subst.  rewrite subst_id; auto. 

assert (J:= wf_tm_gives_wf_env Hwfx).
inversion J; subst.
assert (Jy: wf_env S1 S2 (E& y~:k)).
apply wf_env_cons_var with (k:=k0); auto.

rewrite (@subst_intro x); auto.
(* rewrite (@subst_intro x u); auto.*)
apply simple_tm_substitution with (u:=k); eauto. 
apply typing_weakening; auto.
apply wf_env_cons_var with (k:=k0); auto.  
apply typing_weakening'; auto.

apply wf_tm_var; auto. 
unfold binds_var; auto.
Qed.

Lemma get_atomic_ctrs_sig_unique: forall S L L1 L2,
 get_atomic_ctrs_sig S L = L1 ->
 get_atomic_ctrs_sig S L = L2 ->
 L1 = L2.
Proof.
 (* intros; congruence. *)
 intros S L L1 L2 H J.
 rewrite <- H.
 rewrite <- J.
 reflexivity.
Qed.


Lemma atomic_type_dec: forall S t, decidable (atomic_type S t).
Proof. 
intros.
  remember (get_atomic_ctrs_sig S nil) as L.
  remember (check_simple_type L t) as C. symmetry in HeqC.
  destruct C. 
  left; unfold atomic_type; exists L; auto.
  right; intro J; destruct J as [L' [J1 J2]]. 
  symmetry in HeqL.
  assert (J:= get_atomic_ctrs_sig_unique S nil J1 HeqL). 
  unfold simple_type in J2; subst; auto.
  rewrite J2 in HeqC; inversion HeqC.
Qed.
  
Lemma type_checking_dec: forall S1 S2 E e,
 wf_sig S1 ->
 wf_env S1 S2 E -> 
  sig_extend S1 S2 ->
  term e -> 
  ( decidable (exists t, wf_tm S1 S2 E e t) /\ 
   (forall s args res_tp, all_tm_lc args -> decidable (wf_branches S1 S2 E s args e res_tp))). 
Proof.
 introv Hwf_sig Hwf_env HS Hlc. 
 gen E.
 induction Hlc; intros; split; try solve [right; intro J; inversion J; auto]. 
 Case "fv".
   remember (get a E) as G.
   destruct G.
   destruct b.
  Case "binds_var a t E". 
    symmetry in HeqG. 
     left; exists t; apply wf_tm_var; auto.
   Case "binds_eq".
    right; intro J; destruct J as [t2 Ja]; inversion Ja; subst.
    inversion H2. rewrite H1 in HeqG. inversion HeqG.
  Case "~bind a t E".
  right. intro J. destruct J as [t1 Ja].  inversion Ja; subst. inversion H2. rewrite H1 in HeqG.  inversion HeqG.

 Case "cons".
  destruct (a=== 10).
  Case "con a = K_TYPE".
   subst; left; exists S_KIND; apply wf_tm_K_TYPE; auto.
   destruct (a===11).
   Case "con a = K_PROP".
   subst; left; exists S_KIND; apply wf_tm_K_PROP; auto.
    destruct (a===30).
      Case "con a = T_PRIN".
      subst; left; exists K_TYPE; apply wf_tm_T_PRIN; auto.
        destruct (a===44).
          Case "con a = SELF".
          subst; left; exists T_PRIN; apply wf_tm_TM_SELF; auto.
          Case "otherwise".
           right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; auto.

Case "con". 
   destruct (a === 0).
   subst; left; auto.
   right; intro J; inversion J; auto.

Case "ctr".
  remember (sig_lookup_ctr a S1) as C.
  destruct C.
  Case "sig_lookup_ctr a S = Some t".
  symmetry in HeqC.
  left; exists t; eapply wf_tm_ctr; eauto. 
  Case "sig_lookup_ctr a S = None".
    remember (sig_lookup_assn a S1) as A.
   destruct A. symmetry in HeqA.
   left; exists t; eapply wf_tm_assn; eauto. 
  Case "both lookup fail".
    right. intro J; inversion J. inversion H; subst. inversion H3. rewrite <- HeqC in H2. inversion H2.
    inversion H3. rewrite <- HeqA in H2. inversion H2.
 
Case "app".  
  destruct (IHHlc2 E); auto.  clear H9. destruct H8.
  Case "t2: t". 
   destruct (IHHlc1 E); auto. clear H10. destruct H9.
   Case "t1:t". 
     destruct H9 as [k1 J1].   destruct H8 as [k2 J2].
     assert (Jtm:= is_tm_arr_dec k1 k2).
     destruct Jtm.
     Case "k1 = k2 -> u". 
       destruct H8 as [u Jeq]; subst.
       assert (Jval:= value_dec t2).
       destruct Jval.
       Case "value t2".
       left; exists (u^^t2); eapply wf_tm_app; eauto.
       Case "~value t2".
       assert (Jlc:= lc_dec u).
       destruct Jlc.
       left; exists (u^^t2). eapply wf_tm_app; eauto.
       Case "~value t2 ~lc u".
       right; intro J; destruct J as [ku Jwf]; inversion Jwf; subst; try solve [inversion J1].
       assert (Jeq:= wf_tm_unique Hwf_sig J1 H12).
       inversion Jeq; subst.
       destruct H16; contradiction.
       Case "inversions".
         inversion J1. inversion H17. 
         inversion J1. inversion H18.
         inversion J1. inversion H17.
         inversion J1. inversion H22. inversion H29. inversion H36.
         inversion J1. inversion H18. 
         inversion J1. inversion H17.
     Case "k1 <> k2 -> u".  
      right. intro J3. destruct J3 as [u1 Japp]. 
      inversion Japp; subst; try solve [inversion J1]. 
      assert (Jeq: k1 = tm_arr t0 u). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
      assert (Jeq1: k2 = t0 ). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
      subst; apply H8; exists u; auto. 
       Case "P_SAYS".
       inversion J1. inversion H12. 
       Case "PF_SIGN". inversion J1. inversion H13.
       inversion J1. inversion H12.
       inversion J1. inversion H17. inversion H28. inversion H35.
       inversion J1. inversion H13.
       inversion J1. inversion H12.

   Case "~exists t1:t".
   destruct H8 as [k2 Jwf2].
   right; intro J; destruct J as [ku Jwf]; inversion Jwf; subst.
   apply H9; exists (tm_arr t0 u); auto.
   apply H; exists v; auto.
   apply H0; exists v; auto.
   apply H1; auto.
   apply H2; auto. 
   apply H3; exists e1; auto.
   apply H4; auto. 
   apply H5; exists v1 v2 e1; auto.
   apply H6; exists a; auto.
   apply H7; exists p1; auto.
 
  Case "~exists t, t2:t".
   right; intro; apply H8.
   destruct H9 as [t J].
   inversion J; subst. 
    exists t0; auto.
    exists K_PROP; auto.
    exists K_PROP; auto. 
    rewrite_env (empty & E).
    eapply typing_weakening'; eauto. 
    rewrite_env E; auto. 
    exists K_PROP; auto.
    exists P; auto.
    exists (tm_arr P(tm_app T_PF Q)); auto.
    exists K_PROP; auto.
    exists t; auto.
    exists P; auto.
    exists (tm_arr P(tm_app (tm_app P_SAYS a) Q)); auto.


Case "abs". 
 destruct (IHHlc E); auto. clear H2.  destruct H1.
 Case "t1:t".
 destruct H1 as  [k1 Jwf1].
 pick_fresh z.
 assert (J: z \notin L); auto.
 destruct (H0 z J (E& z~:t1)).  
 eapply wf_env_cons_var; eauto. clear H2. destruct H1. 
 Case "E&z~:t1 |- t2^^x :t".
  destruct H1 as [k2 Jwf2].
  assert (J2:= decompose_open z k2 0).
  destruct J2.
  Case "k2 = u^^z". 
   destruct H1 as [u [Jeq Jz]]; subst.
   clear IHHlc H0.
   assert (Jk1:= is_tm_type_dec k1).
   destruct Jk1.
   Case "k1 = S_KIND \/ k1 = K_TYPE \/ k1 = K_PROP". 
     destruct (wf_tm_tm Hwf_sig Jwf2 HS ).
     Case "u^^z = S_KIND".
       right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
       pick_fresh w.
       assert (Jw: w \notin L0); auto.
       assert (J2:= H5 w Jw). 
       assert (Jeq: u = u2).  eapply open_eq with (k:=0) (x:= w); auto.
       assert (Jwft2: wf_tm S1 S2 (E& w~:t1) (t2^^tm_fv w) (u^^tm_fv w)).
       rewrite (@subst_intro z u); auto.
       eapply typing_rename; eauto. 
       eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
       assert (Jequ: u = S_KIND).  unfold S_KIND. eapply open_eq_const; eauto. 
       inversion H6; subst.
       pick_fresh y.
       assert (Jy: y \notin L1); auto.
       assert (Jinv:= H11  y Jy).
       rewrite Jequ in Jinv. unfold S_KIND in Jinv; unfold open_tm in Jinv ; simpl in Jinv.
       inversion Jinv.

     Case "u^^z: ku".
     destruct H1 as [ku Jwfu].
     assert (Jku := is_tm_type_dec ku). 
     destruct Jku.
     Case "ku = K_TYPE \/ ku = K_PROP \/ ku = S_KIND".
     assert (Jwf_arr: wf_tm S1 S2 E (tm_arr t1 u) ku).
     gather_atoms' as L'.
      eapply wf_tm_arr with (L:= L'); eauto.
      intros.  rewrite <- (@subst_fresh z (tm_fv x) ku).
       apply typing_rename with (x:=z); auto. unfold L' in H2; auto.
        destruct H1. rewrite H1; simpl; auto.
        destruct H1; rewrite H1; simpl; auto.

      destruct (eq_tm_dec ku  S_KIND).
      right; intro J1; destruct J1 as [kt Jwf]; inversion Jwf; subst.
      pick_fresh y.
      assert (Jwft2: wf_tm S1 S2 (E&y~:t1) (t2^^tm_fv y) (u^^tm_fv y)).
      rewrite (@subst_intro z u); auto.
      eapply typing_rename; eauto.
      assert (Ju : u = u2).
       apply open_eq with (k:=0) (x:=y); auto.
       eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.

       subst.
        assert (Jk: k = S_KIND). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
        subst; destruct H8; inversion H2.
      gather_atoms' as L'.
      left; exists (tm_arr t1 u); eapply wf_tm_abs with (L:=L'); eauto.
      intros.       rewrite (@subst_intro z u); auto. 
      eapply typing_rename; eauto. unfold L' in H2; auto.
      destruct H1; auto. destruct H1; auto. contradiction. 
      
       Case "~ ku = K_TYPE \/ ku = K_PROP \/ ku = S_KIND".
       right; intro J1; destruct J1 as [kt Jwf]; inversion Jwf; subst.
       inversion H6; subst.
       pick_fresh y. 
       assert (Jeq: u = u2). 
       apply open_eq with (k:=0) (x:=y); auto.
       assert (Jwfu2: wf_tm S1 S2 (E& y~:t1) (t2^^tm_fv y) (u^^tm_fv y)).
       rewrite   (@subst_intro z u); auto.
       eapply typing_rename; auto. 
       eapply wf_tm_unique with (S1:=S1) (S2:=S2); auto.  apply Jwfu2.
       
       subst. assert (Jeq1: k = ku).
       assert (Jwfu2: wf_tm S1 S2 (E& z~:t1) (u2^^tm_fv z) k).
       rewrite <-  (@subst_fresh y (tm_fv z) k); auto.
       eapply typing_rename; auto.
       apply (wf_tm_unique Hwf_sig Jwfu2 Jwfu); auto.  
       subst; contradiction.
   
   Case "~ k1 = S_KIND \/ k1 = K_TYPE \/ k1 = K_PROP". 
       right; intro J1; apply H0; destruct J1 as [ku Jwf]; inversion Jwf; subst.
       assert (Jeq: k0 = k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
       subst; auto.

  Case "k2<>u^^z".
    right; intro J1; apply H1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
    pick_fresh w.
    assert (Jw: w \notin L0); auto.
    assert (J2:= H5 w Jw). 
    exists u2; eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
    rewrite (@subst_intro w u2); auto.
    eapply typing_rename; eauto. 

 Case "~E&z~:t1 |- t2^^x :t".
  right; intro J1; apply H1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
  pick_fresh w.
  assert (Jw: w \notin L0); auto.
  assert (J2:= H5 w Jw).
  exists (u2^^tm_fv z).
  rewrite (@subst_intro w u2); auto.
  eapply typing_rename; eauto.
 
 Case "~t1:t".
  right; intro J; apply H1; destruct J as [ku Jwf]; inversion Jwf; subst.
  exists k1; auto.

Case "arr". 
 destruct (IHHlc E); auto. clear H2. destruct H1.
 Case "t1:t".
 destruct H1 as  [k1 Jwf1].
 pick_fresh z.
 assert (J: z \notin L); auto. 
 destruct (H0 z J (E& z~:t1)).   
 eapply wf_env_cons_var; eauto. clear H2. destruct H1. 
 Case "t2^^z: t".
 destruct H1 as [k2 Jwf2].
 assert (Jk2:= is_tm_type_dec k2).
 destruct Jk2.
 Case "k2 = K_TYPE \/ k2 = K_PROP \/ k2 = S_KIND".
 gather_atoms' as L'.
 left; exists k2; eapply wf_tm_arr with (L:=L'); eauto.
 intros. rewrite <- (@subst_fresh z (tm_fv x)); auto.
 eapply typing_rename; eauto.
 unfold L' in H2; auto. 
 destruct H1.  rewrite H1; simpl; auto.
 destruct H1; rewrite H1; simpl; auto.
 Case "k2 <>  K_TYPE \/ k2 = K_PROP \/ k2 = S_KIND".
  right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
  pick_fresh y.
  assert (Jwft2: wf_tm S1 S2 (E&z~:t1) (t2^^tm_fv z) ku).
  rewrite <- (@subst_fresh y (tm_fv z) ku); auto.
  eapply typing_rename with (x:=y); eauto.
  assert (Jeq: k2 = ku). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
  rewrite Jeq in H1; contradiction.
  Case "~ t2^^z : t".
    right; intro J1; apply H1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
    pick_fresh y. exists ([y~> (tm_fv z)]ku). eapply typing_rename; eauto.
  Case "~t1: t".
  right; intro J1; apply H1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
  exists k1; auto.

Case "mat". 
 destruct (IHHlc1 E); auto. destruct H.
 Case "t1:t".
 destruct (IHHlc2 E); auto. destruct H1.
 Case "t2: t". 
 destruct H as [k1 Jwft1]. destruct H1 as [k2 Jwft2].
 destruct ( wf_tm_tm Hwf_sig Jwft1 HS).
 Case "k1 = S_KIND". 
 right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst;
 assert (Jeq:= wf_tm_unique Hwf_sig H5 Jwft1); subst;  inversion H6.
 destruct H as [u Jwfk1].
 destruct (eq_tm_dec u K_TYPE). 
 destruct (eq_tm_dec k2 K_TYPE).
  clear IHHlc1 IHHlc2.
  subst.
  assert (J:= fully_applied_args_dec k1).
  destruct J. 
  destruct H as [ctr [args Jfully_app_args]].
  remember (sig_lookup_ctr ctr S1) as C.
  destruct C.
  symmetry in HeqC.
  assert (J:= fully_applied_typ_dec k1 ctr t).
  destruct J.
  destruct (IHHlc3 E); auto.
  assert (Jarg: all_tm_lc (List.rev args)). 
  apply all_tm_lc_list_rev. eapply lc_fully_applied_args_args_lc with (t:= k1); eauto.
  eapply (proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
  assert (J:= H3  k1 (List.rev args) t2 Jarg). 
  destruct J.
  assert (J:= branches_cover_dec S2 t3 ctr).
  destruct J.
  left; exists t2; eapply wf_tm_mat_TYPE; eauto.
  apply HeqC.
  Case "~branches_cover".
  right; intro J1; apply H5; destruct J1 as [ku Jwf]; inversion Jwf; subst.
  assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto. subst.
  destruct (fully_applied_args_unique Jfully_app_args H11). 
  subst; auto.
  assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto. subst.
  destruct (fully_applied_args_unique Jfully_app_args H11). 
  subst; auto.
  
  Case "~wf_branches".
     right; intro J1; apply H4; destruct J1 as [ku Jwf]; inversion Jwf; subst.
     assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
     subst.   destruct (fully_applied_args_unique Jfully_app_args H10). 
     subst; auto.
     assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
     subst.   destruct (fully_applied_args_unique Jfully_app_args H10). 
     subst; auto.
  
  Case "~fully_applied_typ".
       right; intro J1; apply H; destruct J1 as [ku Jwf]; inversion Jwf; subst.
       assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
       subst.   destruct (fully_applied_args_unique Jfully_app_args H7).
       subst.   assert (J1: k = t). eapply wf_ctr_unique; eauto. subst; auto.  
       assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
       subst.   destruct (fully_applied_args_unique Jfully_app_args H7).
       subst.   assert (J1: k = t). eapply wf_ctr_unique; eauto. subst; auto.  
  Case "sig_lookup_ctr = None". 
     right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
     assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
     subst. destruct (fully_applied_args_unique Jfully_app_args H6). 
     subst. inversion H7. rewrite H1 in HeqC. inversion HeqC. 
     assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
     subst. destruct (fully_applied_args_unique Jfully_app_args H6). 
     subst. inversion H7. rewrite H1 in HeqC. inversion HeqC. 
  Case "~ fully_applied".
   right; intro J1; apply H; destruct J1 as [ku Jwf]; inversion Jwf; subst.
   assert (s = k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   subst; exists ctr args; auto.
   assert (s = k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   subst; exists ctr args; auto.

 destruct (eq_tm_dec k2 K_PROP).
 right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
 assert (Jeq:= wf_tm_unique Hwf_sig Jwft2 H10); contradiction.
 assert (J: s = k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
 subst.   assert (Jeq:= wf_tm_unique Hwf_sig H5 Jwfk1); contradiction.
 
 Case "t2 not K_TYPE or K_PROP".
 right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
 assert (J: k2 = K_TYPE). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto. contradiction.
  assert (J: k2 = K_PROP). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto. contradiction.
 

 destruct (eq_tm_dec u K_PROP).
  destruct (eq_tm_dec k2 K_PROP).
  clear IHHlc1 IHHlc2.
  subst.
  assert (J:= fully_applied_args_dec k1).
  destruct J. 
  destruct H as [ctr [args Jfully_app_args]].
  remember (sig_lookup_ctr ctr S1) as C.
  destruct C.
  symmetry in HeqC.
  assert (J:= fully_applied_typ_dec k1 ctr t).
  destruct J.
  destruct (IHHlc3 E); auto.
  assert (Jarg: all_tm_lc (List.rev args)). 
    apply all_tm_lc_list_rev. eapply lc_fully_applied_args_args_lc with (t:= k1); eauto.
  eapply (proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
  assert (J:= H3  k1 (List.rev args) t2 Jarg). 
  destruct J.
  assert (J:= branches_cover_dec S2 t3 ctr).
  destruct J.
  left; exists t2; eapply wf_tm_mat_PROP; eauto.
  apply HeqC.
  Case "~branches_cover".
  right; intro J1; apply H5; destruct J1 as [ku Jwf]; inversion Jwf; subst.
  assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto. subst.
  destruct (fully_applied_args_unique Jfully_app_args H11). 
  subst; auto.
  assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto. subst.
  destruct (fully_applied_args_unique Jfully_app_args H11). 
  subst; auto.
  
  Case "~wf_branches".
     right; intro J1; apply H4; destruct J1 as [ku Jwf]; inversion Jwf; subst.
     assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
     subst.   destruct (fully_applied_args_unique Jfully_app_args H10). 
     subst; auto.
     assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
     subst.   destruct (fully_applied_args_unique Jfully_app_args H10). 
     subst; auto.
  
  Case "~fully_applied_typ".
       right; intro J1; apply H; destruct J1 as [ku Jwf]; inversion Jwf; subst.
       assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
       subst.   destruct (fully_applied_args_unique Jfully_app_args H7).
       subst.   assert (J1: k = t). eapply wf_ctr_unique; eauto. subst; auto.  
       assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
       subst.   destruct (fully_applied_args_unique Jfully_app_args H7).
       subst.   assert (J1: k = t). eapply wf_ctr_unique; eauto. subst; auto.  
  Case "sig_lookup_ctr = None". 
     right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
     assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
     subst. destruct (fully_applied_args_unique Jfully_app_args H6). 
     subst. inversion H7. rewrite H1 in HeqC. inversion HeqC. 
     assert (J1: s= k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
     subst. destruct (fully_applied_args_unique Jfully_app_args H6). 
     subst. inversion H7. rewrite H1 in HeqC. inversion HeqC. 
  Case "~ fully_applied".
   right; intro J1; apply H; destruct J1 as [ku Jwf]; inversion Jwf; subst.
   assert (s = k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   subst; exists ctr args; auto.
   assert (s = k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   subst; exists ctr args; auto.

 destruct (eq_tm_dec k2 K_TYPE).
 right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
 assert (J: s = k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
 subst.   assert (Jeq:= wf_tm_unique Hwf_sig Jwfk1 H5); contradiction.
 assert (Jeq:= wf_tm_unique Hwf_sig Jwft2 H10); contradiction.
 
 Case "t2 not K_TYPE or K_PROP".
 right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
 assert (J: k2 = K_TYPE). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto. contradiction.
  assert (J: k2 = K_PROP). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto. contradiction.
 
 Case "k1 is not K_TYPE or K_PROP".
 right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
  assert (J: s = k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
  subst.   assert (J:= wf_tm_unique Hwf_sig Jwfk1 H5); contradiction.
  assert (J: s = k1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
  subst.   assert (J:= wf_tm_unique Hwf_sig Jwfk1 H5); contradiction.

 Case "~t2:t".
   right; intro J1; apply H1; destruct J1 as [ku Jwf]; inversion Jwf; subst.
   exists K_TYPE; auto. exists K_PROP; auto.
 Case "~t1:t".
  right; intro J1; apply H; destruct J1 as [ku Jwf]; inversion Jwf; subst; exists s; auto.

Case "branch". 
     right; intro J1; destruct J1 as [ku Jwf]; inversion Jwf.
Case "branch".
 introv Hargs.
 destruct (IHHlc1 E); auto. 
 destruct (IHHlc2 E); auto. destruct (H2 s args res_tp); auto.
 remember (sig_lookup_ctr c S1) as C.
  destruct C.
  symmetry in HeqC.
 clear H0 H1 IHHlc1 IHHlc2.
 destruct H. destruct H as [body_tp Jwft2].
  assert (J: decidable (wf_brn S1 S2 s args  t body_tp res_tp)).
  eapply wf_brn_dec; eauto.
  assert (Jwfc: wf_tm S1 S2 E (tm_ctr c) t); auto.
  destruct (wf_tm_tm Hwf_sig Jwfc HS ).
  subst. unfold S_KIND; simpl; auto. destruct H.
  eapply (proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto. 
  destruct J. left; eapply wf_branches_Some; eauto. apply HeqC.
  Case "~wf_brn".
   right; intro J; apply H; inversion J; subst; eauto. 
   assert (Jeq: t = ectr_typ). eapply wf_ctr_unique; eauto. apply HeqC.
  assert (Jeq': body_tp = body_type).  eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.  subst; auto.
 Case "~t2:t". 
   right; intro J; apply H; inversion J; subst; eauto.
 Case "sig_lookup_ctr c S = None".
     right; intro J; inversion J; subst; eauto.
     inversion H11. rewrite H5 in HeqC. inversion HeqC.
 Case "~wf_branches rest".
  right; intro J; apply H3; inversion J; subst; auto.

Case "cast".
  destruct (IHHlc1 E); auto. clear H0. destruct H.
  Case "t1:t".      
   destruct (IHHlc2 E); auto. clear H1. destruct H0.
   Case "t2:k2". 
     destruct H as [k1 J1]. destruct H0 as [k2 J2].
     destruct (eq_tm_dec k2 K_TYPE).
     Case "k2 = KYPE".
     assert (J3:= converts_dec k1 t2 Hwf_env).
     destruct J3. subst.
     left; exists t2.  apply wf_tm_cast with (t1:=k1); auto.
      
     right; intro J; apply H.
     destruct J as [k Jwf]. inversion Jwf; subst.
      assert (Jeq: k1 = t0). eapply wf_tm_unique with (S1:=S1) (S2:= S2); eauto. 
      subst; auto.
     Case "k2 <>K_TYPE".
     right; intro J; apply n.
     destruct J as [k Jwf]. inversion Jwf; subst. eapply wf_tm_unique with (S1:=S1) (S2:=S2) ; eauto. 
   Case "~t2:k2".
    right; intro J; apply H0.
    destruct J as [k Jwf]. inversion Jwf; subst. exists K_TYPE; auto.
  Case "~t1:t".
   right; intro J; apply H.
   destruct J as [k Jwf]. inversion Jwf; subst. exists t0; auto.

Case "P_SAYS". 
 destruct (IHHlc1 E); auto. clear H0. destruct H. 
 destruct (IHHlc2 E); auto. clear H1. destruct H0.
 destruct H as [k1 Jwft1].  destruct H0 as [k2 Jwft2].
 destruct (eq_tm_dec k1 T_PRIN).
 destruct (eq_tm_dec k2 K_PROP).
 destruct (value_dec t1).
 subst; left; exists K_PROP; eapply wf_tm_P_SAYS; eauto.
  right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst; auto.
  inversion H2. inversion H7. 
  right; intro J; destruct J as [ku Jwf]; inversion Jwf; subst.
  inversion H1. inversion H6. assert (J:= wf_tm_unique Hwf_sig Jwft2 H3); contradiction.
  right; intro J; destruct J as [ku Jwf]; inversion Jwf; subst.
  inversion H1. inversion H6. assert (J:= wf_tm_unique Hwf_sig Jwft1 H1); contradiction.
  right; intro J; apply H0; destruct J as [k1 Jwf1]; inversion Jwf1; subst.
  inversion H3. inversion H8. exists K_PROP; auto.
  right; intro J; apply H; destruct J as [k1 Jwf1]; inversion Jwf1; subst.
  inversion H2. inversion H7. exists T_PRIN; auto.

Case "PF_SIGN".
  destruct (IHHlc1 empty); auto.  clear H0. destruct H. 
  destruct (IHHlc2 empty); auto. clear H1. destruct H0.
  destruct H as [k1 Jwft1].  destruct H0 as [k2 Jwft2].
  destruct (eq_tm_dec k1 T_PRIN).
  destruct (eq_tm_dec k2 K_PROP).
  destruct (value_dec t1).
  subst; left; exists (tm_app (tm_app P_SAYS t1) t2); eapply wf_tm_PF_SIGN; eauto.
  right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst; auto.
  inversion H2. inversion H7. 
  right; intro J; destruct J as [ku Jwf]; inversion Jwf; subst.
  inversion H1. inversion H6. assert (J:= wf_tm_unique Hwf_sig Jwft2 H4); contradiction.
  right; intro J; destruct J as [ku Jwf]; inversion Jwf; subst.
  inversion H1. inversion H6. assert (J:= wf_tm_unique Hwf_sig Jwft1 H1); contradiction.
  right; intro J; apply H0; destruct J as [k1 Jwf1]; inversion Jwf1; subst.
  inversion H3. inversion H8. exists K_PROP; auto.
  right; intro J; apply H; destruct J as [k1 Jwf1]; inversion Jwf1; subst.
  inversion H2. inversion H7. exists T_PRIN; auto.

Case "T_PF".
   destruct (IHHlc E); auto. clear H0. destruct H.
   destruct H as [k Ht].
   destruct (eq_tm_dec k K_PROP).
   subst; left; exists K_TYPE; apply wf_tm_T_PF; auto.
   subst; right; intro J; destruct J as [k1 Jwf1]; inversion Jwf1; subst. 
   inversion H1.
   assert (Jeq := wf_tm_unique Hwf_sig Ht H1); contradiction.
   subst; right; intro J; apply H; destruct J as [k1 Jwf1]; inversion Jwf1; subst.
   exists t2; auto.
   exists K_PROP; auto.

Case "TM_PF_RETRUN".
    destruct (IHHlc E); auto. clear H0. destruct H.
    destruct H as [k Ht].
    assert (J:= wf_tm_tm Hwf_sig  Ht HS).
    destruct J.
     right; subst; intro J; destruct J as [k Jwf]; inversion Jwf; subst.
     inversion H1. assert (J1:= wf_tm_unique Hwf_sig H0 Ht); subst; inversion H2.
     
     destruct H as [ku Jwf_k2].
     destruct (eq_tm_dec ku K_PROP).
     left; subst; exists (tm_app T_PF k); apply wf_tm_TM_PF_RETURN; auto.
     right; subst; intro J; destruct J as [k3 Jwf3]; inversion Jwf3; subst.
     inversion H1.  
     assert (Jeq1:= wf_tm_unique Hwf_sig H0 Ht); subst.
     assert (Jeq2:= wf_tm_unique Hwf_sig Jwf_k2 H2); contradiction.

     right; intro J; apply H; destruct J as [k1 Jwf1]; inversion Jwf1; subst.
     inversion H2. exists P; auto.

Case "TM_PF_BIND".
   destruct (IHHlc1 E); auto. clear H0. destruct H.
   destruct (IHHlc2 E); auto. clear H1. destruct H0.
   destruct H as [k1 Jwft1]. destruct H0 as [k2 Jwft2].
   destruct (is_tm_app_dec k1). destruct H as [e1 [P Jwfk1]].
   destruct (eq_tm_dec e1 T_PF).
   destruct (is_tm_arr_dec k2 P).
   destruct H as [ku Jk2].
   destruct (is_tm_app_dec ku). 
   destruct H as [u1 [Q Jku]]. 
   destruct (eq_tm_dec u1 T_PF). 
   destruct (lc_dec Q).
   subst; left; exists (tm_app T_PF Q); eapply wf_tm_TM_PF_BIND  with (P:=P); eauto.
   Case "~lc Q".
   subst; right; intro J; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq1:= wf_tm_unique Hwf_sig Jwft1 H2). inversion Jeq1; subst.
   assert (Jeq: tm_arr P0 (tm_app T_PF Q) = tm_arr P0 (tm_app T_PF Q0)). 
    eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   inversion Jeq; subst; apply H; auto.
   Case "k1<>T_PF u".
   subst; right; intro J; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H1. inversion H6.
   assert (Jeq1:= wf_tm_unique Hwf_sig Jwft2 H3). inversion Jeq1; subst.
   apply n; auto.
   Case "k2<> P-> u1 u2".
   subst; right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq1:= wf_tm_unique Hwf_sig Jwft2 H4). inversion Jeq1; subst; eauto.

   Case "k2<> P-> u".
   right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq1:= wf_tm_unique Hwf_sig Jwft1 H2). inversion Jeq1; subst.
   assert (Jeq:= wf_tm_unique Hwf_sig Jwft2 H4). inversion Jeq; subst; eauto.
   Case "k1 <>T_PF P".
   right; intro J; apply n; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H1. inversion H6.
   assert (Jeq:= wf_tm_unique Hwf_sig H1 Jwft1). inversion Jeq; auto.
   Case "k1<> u1 u2".
   right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq:= wf_tm_unique Hwf_sig Jwft1 H2); eauto.
   Case "~t2:k2".
   subst; right; intro J; apply H0; destruct J as [k1 Jwf1]; inversion Jwf1; subst; eauto.
   Case "~t1:k1".
   subst; right; intro J; apply H; destruct J as [k1 Jwf1]; inversion Jwf1; subst; eauto.
   inversion H2. inversion H7. 

Case "TM_SAY".
    destruct (IHHlc E); auto. clear H0. destruct H.
    destruct H as [k Ht].    
    destruct (eq_tm_dec k K_PROP).
    subst; left; exists (tm_app T_PF (tm_app (tm_app P_SAYS TM_SELF) t)); apply wf_tm_TM_SAY; auto.
    subst; right; intro J; destruct J as [k1 Jwf1]; inversion Jwf1; subst. 
    inversion H1. assert (Jeq := wf_tm_unique Hwf_sig Ht H1); contradiction.

     right; intro J; apply H; destruct J as [k1 Jwf1]; inversion Jwf1; subst; eauto.

Case "TM_PIF". 
  destruct (IHHlc1 E); auto. clear H0. destruct H. destruct H as [k1 Jwft1].
  destruct (IHHlc2 E); auto. clear H0. destruct H. destruct H as [k2 Jwft2]. 
  destruct (eq_tm_dec k1 k2). subst.
  destruct (atomic_type_dec S2 k2 ) . 
  destruct (value_dec t1). 
  destruct (value_dec t2).
  assert (J:= wf_tm_tm Hwf_sig Jwft1 HS).
  destruct J. subst. unfold atomic_type in H. 
  destruct H as [L [Jget Jsimpl]]. inversion Jsimpl.
  destruct H2 as [K Jwfk2].
  destruct (eq_tm_dec K K_TYPE). 
  pick_fresh z.
  destruct (IHHlc3 (E& z~(bnd_eq t1 t2 k2))); auto.
  apply wf_env_cons_eq; auto. rewrite e in Jwfk2; auto.  
   clear H3.
    destruct H2. destruct H2 as [k3 Jwft3].
  destruct (IHHlc4 E); auto. clear H3. destruct H2. destruct H2 as [k4 Jwft4].
  destruct (eq_tm_dec k3 k4).
(*
  destruct (wf_tm_tm Hwf_sig Jwft1); auto.
  subst; right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; eauto.
  inversion H. destruct H2. inversion H3.    inversion H. destruct H2. inversion H3. 
  destruct H2 as [K Jwfk2].
  destruct (eq_tm_dec K K_TYPE) *)
  gather_atoms' as L'.
  subst; left; exists k4; eapply wf_tm_TM_PIF with (L:=L'); eauto.
  intros.  eapply typing_eq_rename; eauto. unfold L' in H2; auto. 

  Case "k3<>k4".
  right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; apply n; eauto.
  inversion H4. inversion H9. inversion H16. inversion H23.
  assert (Jeq:= wf_tm_unique Hwf_sig Jwft4 H14). subst.
  assert (Jeq1: k2 = T1). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto. subst.
  pick_fresh y. 
  assert (J4: wf_tm S1 S2 (E&z~bnd_eq t1 t2 T1) t3 k). 
  eapply typing_eq_rename with (x:=y); eauto. 
  eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
  Case "~t4:t".
  right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; apply H2; eauto.
  Case "~t3:t".
  right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; apply H2; eauto.
  inversion H5. inversion H10. inversion H17. inversion H24.
  pick_fresh y. 
  exists k.
  assert (Jeq: T1 = k2).  eapply wf_tm_unique; eauto. subst.
    eapply typing_eq_rename with (x:=y); eauto. 
  Case "K<>K_TYPE".
    subst; right; intro J; apply n; destruct J as [k Jwf]; inversion Jwf; subst; eauto.
    inversion H4. inversion H9. inversion H16. inversion H23.
    assert (J: k2 = T1).  eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   subst; eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
    
  Case "~value t2".
  right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; apply H1; eauto.
  inversion H4. inversion H9. inversion H16. inversion H23.
  Case "~value t1".
  right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; apply H0; eauto.
  inversion H3. inversion H8. inversion H15. inversion H22.
  Case "~atomic_type k2".
  right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; apply H; eauto.
  inversion H2. inversion H7. inversion H14. inversion H21.
  assert (J:= wf_tm_unique Hwf_sig H4 Jwft1); subst; auto.
  Case "k1<>k2".
  right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; apply n; eauto.
  inversion H1. inversion H6. inversion H13. inversion H20.
  assert (Jeq:= wf_tm_unique Hwf_sig Jwft1 H3); subst.
  eapply wf_tm_unique with (S1:=S1) (S2:=S2) (e:=t2); eauto.
  Case "~t2:t".
  right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; apply H; eauto.
  inversion H2. inversion H7. inversion H14. inversion H21.
  Case "~t1:t".
  right; intro J; destruct J as [k Jwf]; inversion Jwf; subst; apply H; eauto.
  inversion H2. inversion H7. inversion H14. inversion H21.


Case "PF_SAYS_RETURN".
    destruct (IHHlc1 E); auto. clear H0. destruct H.
    destruct H as [k1 Jwft1].
    destruct (IHHlc2 E); auto. clear H0. destruct H.
    destruct H as [k2 Jwft2].
   
    assert (J:= wf_tm_tm Hwf_sig Jwft2 HS).
    destruct J.
     right; subst; intro J; destruct J as [k Jwf]; inversion Jwf; subst.
     inversion H1. inversion H6. assert (J1:= wf_tm_unique Hwf_sig H1 Jwft2); subst. inversion H2.
     
     destruct H as [ku Jwf_k].
     destruct (eq_tm_dec ku K_PROP).
     destruct (eq_tm_dec k1 T_PRIN).
     destruct (value_dec t1).
     left; subst; exists (tm_app (tm_app P_SAYS t1) k2); apply wf_tm_PF_SAYS_RETURN; auto.

     right; subst; intro J; apply H; destruct J as [k3 Jwf3]; inversion Jwf3; subst; auto.
     inversion H2. inversion H7. 
     right; subst; intro J; destruct J as [k3 Jwf3]; inversion Jwf3; subst.
     inversion H1. inversion H6.  
     assert (Jeq1:= wf_tm_unique Hwf_sig Jwft1 H4); contradiction.
     right; subst; intro J; destruct J as [k3 Jwf3]; inversion Jwf3; subst.
     inversion H1. inversion H6. 
     assert (Jeq: k2 = P). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
     subst; assert (Jeq2:= wf_tm_unique Hwf_sig Jwf_k H2); contradiction.
     right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst; eauto.
     right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst; eauto.
     inversion H2. inversion H7. 

Case "PF_SAYS_BIND".
   destruct (IHHlc1 E); auto. clear H0. destruct H.
   destruct (IHHlc2 E); auto. clear H1. destruct H0. 
   destruct H as [k1 Jwft1]. destruct H0 as [k2 Jwft2].
   destruct (is_tm_app_dec k1).    destruct H as [u1 [u2 Jk1]]. 
   destruct (is_tm_app_dec u1).    destruct H as [s1 [s2 Ju1]].
   destruct (eq_tm_dec s1 P_SAYS).  subst.
   destruct (is_tm_arr_dec k2 u2).   destruct H as [ku Jk2].
   destruct (is_tm_app_dec ku).   destruct H as [u1 [Q Jku]].
   destruct (is_tm_app_dec u1). destruct H as [ b2 [v Ju1]].
   destruct (eq_tm_dec b2 P_SAYS).  
   destruct (lc_dec Q). subst.
   destruct (eq_tm_dec s2 v).
   subst; left; exists (tm_app (tm_app P_SAYS v) Q); eapply wf_tm_PF_SAYS_BIND; eauto.
   Case "s2<>v".
   subst; right; intro J; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq1:= wf_tm_unique Hwf_sig Jwft1 H2). inversion Jeq1; subst.
   assert (Jeq: tm_arr P (tm_app (tm_app P_SAYS a) Q0) = tm_arr P (tm_app (tm_app P_SAYS v) Q)).
   eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   inversion Jeq; subst. apply n; auto.
   Case "~lc Q".
   right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq: tm_arr P (tm_app (tm_app P_SAYS a) Q0) = tm_arr u2 (tm_app (tm_app P_SAYS v) Q)). 
   eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   inversion Jeq; subst; auto.
   Case "ku<>P_SAYS x y".
   subst; right; intro J; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H1. inversion H6.
   assert (Jeq: tm_arr P (tm_app (tm_app P_SAYS a) Q0) = tm_arr u2 (tm_app (tm_app b2 v) Q)). 
   eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   inversion Jeq; subst; auto.

   Case "k2 <> P-> x y".
   subst; right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq:= wf_tm_unique Hwf_sig Jwft2 H4).
   inversion Jeq; subst. exists P_SAYS a; auto.

   Case "k2<> P -> (u1 u2)".
   right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq1:= wf_tm_unique Hwf_sig Jwft1 H2). inversion Jeq1; subst.
   assert (Jeq: tm_arr P ku = tm_arr P (tm_app (tm_app P_SAYS a) Q)). 
    eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   inversion Jeq; subst. exists (tm_app P_SAYS a) Q; auto.
   Case "k2<> u2-> u".
   right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq1:= wf_tm_unique Hwf_sig Jwft1 H2). inversion Jeq1; subst.
   assert (Jeq: k2 = tm_arr P (tm_app (tm_app P_SAYS a) Q)). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   exists (tm_app (tm_app P_SAYS a) Q); auto.
   Case "k1 <>P_SAYS s2 u2".
   subst; right; intro J; apply n; destruct J as [k Jwf]; inversion Jwf; subst; auto.
   inversion H1. inversion H6.
   assert (Jeq: tm_app (tm_app s1 s2) u2 = tm_app (tm_app P_SAYS a) P). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   inversion Jeq; auto.
   Case "k1<> (s1 s2) u2".
   right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq: tm_app u1 u2 = tm_app (tm_app P_SAYS a) P). 
   eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   inversion Jeq; subst.  exists P_SAYS a; auto.
   Case "k1<> u1 u2".
   right; intro J; apply H; destruct J as [k Jwf]; inversion Jwf; subst.
   inversion H2. inversion H7.
   assert (Jeq: k1 = tm_app (tm_app P_SAYS a) P). eapply wf_tm_unique with (S1:=S1) (S2:=S2); eauto.
   exists (tm_app P_SAYS a) P; auto.
   Case "~t2:k2".
   subst; right; intro J; apply H0; destruct J as [k1 Jwf1]; inversion Jwf1; subst.
   inversion H3. inversion H8. exists (tm_arr P(tm_app (tm_app P_SAYS a) Q)); auto.
   Case "~t1:k1".
   subst; right; intro J; apply H; destruct J as [k1 Jwf1]; inversion Jwf1; subst.
   inversion H2. inversion H7. exists (tm_app (tm_app  P_SAYS a) P); auto.
  
Qed.


(***********************************************************************)
(** term dec *)
Hint Constructors term.

Definition P_term__lc (n:nat) := forall t, tm_size t = n -> 
  term t -> lc t.

Lemma _term__lc: forall n, P_term__lc(n).
intros.
  apply lt_wf_ind. clear n.
  intros n H.
  unfold P_term__lc in *.
  unfold decidable in *.
  destruct t; intros; inversion H1; subst; auto.

  assert (tm_size t1 < tm_size (tm_app t1 t2)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_app t1 t2)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H2; auto.

  assert (tm_size t2 < tm_size (tm_app (tm_app P_SAYS t0) t2)). simpl. omega.
  assert (tm_size t0 < tm_size (tm_app (tm_app P_SAYS t0) t2)). simpl. omega.
  apply H with (t:=t2) in H0; auto.
  apply H with (t:=t0) in H2; auto.
  apply lc_app; auto.
    apply lc_app; auto.
      unfold P_SAYS. apply lc_con.

  assert (tm_size t2 < tm_size (tm_app (tm_app PF_SIGN t0) t2)). simpl. omega.
  assert (tm_size t0 < tm_size (tm_app (tm_app PF_SIGN t0) t2)). simpl. omega.
  apply H with (t:=t2) in H0; auto.
  apply H with (t:=t0) in H2; auto.
  apply lc_app; auto.
    apply lc_app; auto.
      unfold PF_SIGN. apply lc_con.

  assert (tm_size t2 < tm_size (tm_app T_PF t2)). simpl. omega.
  apply H with (t:=t2) in H0; auto.
  apply lc_app; auto.
      unfold T_PF. apply lc_con.

  assert (tm_size t2 < tm_size (tm_app TM_PF_RETURN t2)). simpl. omega.
  apply H with (t:=t2) in H0; auto.
  apply lc_app; auto.
      unfold TM_PF_RETURN. apply lc_con.

  assert (tm_size t2 < tm_size (tm_app (tm_app TM_PF_BIND t0) t2)). simpl. omega.
  assert (tm_size t0 < tm_size (tm_app (tm_app TM_PF_BIND t0) t2)). simpl. omega.
  apply H with (t:=t2) in H0; auto.
  apply H with (t:=t0) in H2; auto.
  apply lc_app; auto.
    apply lc_app; auto.
      unfold TM_PF_BIND. apply lc_con.

  assert (tm_size t2 < tm_size (tm_app TM_SAY t2)). simpl. omega.
  apply H with (t:=t2) in H0; auto.
  apply lc_app; auto.
      unfold TM_SAY. apply lc_con.

  assert (tm_size t2 < tm_size (tm_app (tm_app (tm_app (tm_app TM_PIF t0) t3) t4) t2)). simpl. omega.
  assert (tm_size t4 < tm_size (tm_app (tm_app (tm_app (tm_app TM_PIF t0) t3) t4) t2)). simpl. omega.
  assert (tm_size t3 < tm_size (tm_app (tm_app (tm_app (tm_app TM_PIF t0) t3) t4) t2)). simpl. omega.
  assert (tm_size t0 < tm_size (tm_app (tm_app (tm_app (tm_app TM_PIF t0) t3) t4) t2)). simpl. omega.
  apply H with (t:=t2) in H0; auto.
  apply H with (t:=t4) in H2; auto.
  apply H with (t:=t3) in H3; auto.
  apply H with (t:=t0) in H8; auto.
  apply lc_app; auto.
  apply lc_app; auto.
  apply lc_app; auto.
  apply lc_app; auto.
      unfold TM_PIF. apply lc_con.

  assert (tm_size t2 < tm_size (tm_app (tm_app PF_SAYS_RETURN t0) t2)). simpl. omega.
  assert (tm_size t0 < tm_size (tm_app (tm_app PF_SAYS_RETURN t0) t2)). simpl. omega.
  apply H with (t:=t2) in H0; auto.
  apply H with (t:=t0) in H2; auto.
  apply lc_app; auto.
    apply lc_app; auto.
      unfold PF_SAYS_RETURN. apply lc_con.

  assert (tm_size t2 < tm_size (tm_app (tm_app PF_SAYS_BIND t0) t2)). simpl. omega.
  assert (tm_size t0 < tm_size (tm_app (tm_app PF_SAYS_BIND t0) t2)). simpl. omega.
  apply H with (t:=t2) in H0; auto.
  apply H with (t:=t0) in H2; auto.
  apply lc_app; auto.
    apply lc_app; auto.
      unfold PF_SAYS_BIND. apply lc_con.

  assert (tm_size t1 < tm_size (tm_abs t1 t2)). simpl. omega.
  pick_fresh y.
  assert (tm_size (t2^^(tm_fv y)) < tm_size (tm_abs t1 t2)).
    simpl. unfold open_tm. rewrite open_tm_size_eq. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2^^(tm_fv y)) in H2; auto.
    apply lc_abs with (L:=fv t1 \u fv t2); auto.
    intros. apply swap_lc with (x:=y); auto.

  assert (tm_size t1 < tm_size (tm_abs t1 t2)). simpl. omega.
  pick_fresh y.
  assert (tm_size (t2^^(tm_fv y)) < tm_size (tm_abs t1 t2)).
    simpl. unfold open_tm. rewrite open_tm_size_eq. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2^^(tm_fv y)) in H2; auto.
    apply lc_arr with (L:=fv t1 \u fv t2); auto.
    intros. apply swap_lc with (x:=y); auto.

  assert (tm_size t1 < tm_size (tm_mat t1 t2 t3)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_mat t1 t2 t3)). simpl. omega.
  assert (tm_size t3 < tm_size (tm_mat t1 t2 t3)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H2; auto.
  apply H with (t:=t3) in H3; auto.

  assert (tm_size t1 < tm_size (tm_brn c t1 t2)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_brn c t1 t2)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H2; auto.

  assert (tm_size t1 < tm_size (tm_cast t1 t2)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_cast t1 t2)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H2; auto.
Qed.

Lemma term__lc: forall t, term t -> lc t.
intros.
  assert (P_term__lc (tm_size t)).
    apply (@_term__lc (tm_size t)).
  unfold P_term__lc in H0.
  eapply H0; eauto.
Qed.

Lemma notin_fv_open : forall Y X T k,
  X \notin fv T ->  X \notin {{Y}} ->
  X \notin fv (open_tm_aux k (tm_fv Y) T)
 .
Proof.
 intros. gen Y X k.
 induction T; intros; unfold open_tm in *; simpl in *; notin_simpl; auto.
    destruct (b===k); simpl; auto.
Qed.    

Definition P_subst_fix (n:nat) := forall t (x z:atom) c, 
  tm_size t = n -> 
  x <> z -> 
  term (tm_fv z) -> term t -> 
  ~ (exists v : tm, t = tm_app (tm_con c) v) ->
  ~ (exists v : tm, [x~>(tm_fv z)]t = tm_app (tm_con c) v) .

Lemma _subst_fix: forall n, P_subst_fix(n).
intros.
  apply lt_wf_ind. clear n.
  intros n H.
  unfold P_subst_fix in *.
  destruct t; intros;
    try solve [simpl; unfold not; intros; inversion H5; inversion H6].
    simpl; destruct (a==x); unfold not; intros; inversion H5; inversion H6.
    destruct t1; intros;
      try solve [simpl; unfold not; intros; inversion H5; inversion H6].
      simpl; destruct (a==x); unfold not; intros; inversion H5; inversion H6.
      destruct (eq_tm_dec (tm_con c0) (tm_con c)).
        inversion e. subst.
        assert (False).
          apply H4. apply ex_intro with (x:=t2). reflexivity.
        inversion H0.

        simpl. unfold not; intros; inversion H5; inversion H6.  subst. apply n0; auto.
Qed.

Lemma subst_fix:  forall t (x z:atom) c, 
  x <> z -> 
  term (tm_fv z) -> term t -> 
  ~ (exists v : tm, t = tm_app (tm_con c) v) ->
  ~ (exists v : tm, [x~>(tm_fv z)]t = tm_app (tm_con c) v) .
intros.
  assert (P_subst_fix (tm_size t)).
    apply (@_subst_fix (tm_size t)).
  unfold P_subst_fix in H3.
  eapply H3; eauto.
Qed.

Definition P_subst_fix3 (n:nat) := forall t (x z:atom) c, 
  tm_size t = n -> 
  x <> z -> 
  term (tm_fv z) -> term t -> 
  ~ (exists v1 : tm,
         exists v2 : tm,
            exists v3 : tm,
               t = tm_app (tm_app (tm_app (tm_con c) v1) v2) v3)
  ->
  ~ (exists v1 : tm,
         exists v2 : tm,
            exists v3 : tm,
               [x~>(tm_fv z)]t = tm_app (tm_app (tm_app (tm_con c) v1) v2) v3)
  .

Lemma _subst_fix3: forall n, P_subst_fix3(n).
intros.
  apply lt_wf_ind. clear n.
  intros n H.
  unfold P_subst_fix3 in *.
  destruct t; intros;
    try solve [simpl; unfold not; intros; inversion H5; inversion H6; inversion H7; inversion H8].
    simpl; destruct (a==x); unfold not; intros; inversion H5; inversion H6; inversion H7; inversion H8.
  destruct t1; intros;
    try solve [simpl; unfold not; intros; inversion H5; inversion H6; inversion H7; inversion H8].
    simpl; destruct (a==x); unfold not; intros; inversion H5; inversion H6; inversion H7; inversion H8.
  destruct t1_1; intros;
    try solve [simpl; unfold not; intros; inversion H5; inversion H6; inversion H7; inversion H8].
    simpl; destruct (a==x); unfold not; intros; inversion H5; inversion H6; inversion H7; inversion H8.
  destruct t1_1_1; intros;
    try solve [simpl; unfold not; intros; inversion H5; inversion H6; inversion H7; inversion H8].
    simpl; destruct (a==x); unfold not; intros; inversion H5; inversion H6; inversion H7; inversion H8.
  destruct (eq_tm_dec (tm_con c0) (tm_con c)).
    inversion e. subst.
    assert (False).
      apply H4. 
      apply ex_intro with (x:=t1_1_2). 
      apply ex_intro with (x:=t1_2). 
      apply ex_intro with (x:=t2). 
      reflexivity.
    inversion H0.

    simpl. unfold not; intros; inversion H5; inversion H6; inversion H7; inversion H8. subst. apply n0; auto.
Qed.

Lemma subst_fix3:  forall t (x z:atom) c, 
  x <> z -> 
  term (tm_fv z) -> term t -> 
  ~ (exists v1 : tm,
         exists v2 : tm,
            exists v3 : tm,
               t = tm_app (tm_app (tm_app (tm_con c) v1) v2) v3)
  ->
  ~ (exists v1 : tm,
         exists v2 : tm,
            exists v3 : tm,
               [x~>(tm_fv z)]t = tm_app (tm_app (tm_app (tm_con c) v1) v2) v3)
  .
intros.
  assert (P_subst_fix3 (tm_size t)).
    apply (@_subst_fix3 (tm_size t)).
  unfold P_subst_fix3 in H3.
  eapply H3; eauto.
Qed.

Lemma subst_term : forall (t:tm) (x z:atom),
  x <> z -> 
  term (tm_fv z) -> term t -> 
  term ([x~>(tm_fv z)]t).
Proof.
  induction 3; intros; simpl in *; auto*.
  destruct (a== x); auto*.

  apply term_app; auto. 
   unfold P_SAYS. apply subst_fix; auto.
   unfold PF_SIGN. apply subst_fix; auto.
   unfold T_PF. destruct t1;
     try solve [simpl; unfold not; intro; inversion H10].
     simpl. destruct (a==x); auto.
                   unfold not; intro; inversion H10.
     simpl. unfold not; intro; inversion H10.
       subst. unfold T_PF in H3. apply H3; auto.
   unfold TM_PF_RETURN. destruct t1;
     try solve [simpl; unfold not; intro; inversion H10].
     simpl. destruct (a==x); auto.
                   unfold not; intro; inversion H10.
     simpl. unfold not; intro; inversion H10.
       subst. unfold TM_PF_RETURN in H4. apply H4; auto.
   unfold TM_PF_BIND. apply subst_fix; auto.
   unfold TM_SAY. destruct t1;
     try solve [simpl; unfold not; intro; inversion H10].
     simpl. destruct (a==x); auto.
                   unfold not; intro; inversion H10.
     simpl. unfold not; intro; inversion H10.
       subst. unfold TM_SAY in H6. apply H6; auto.
   unfold TM_PIF. apply subst_fix3; auto.
   unfold PF_SAYS_RETURN. apply subst_fix; auto.
   unfold PF_SAYS_BIND. apply subst_fix; auto.

  apply_fresh term_abs as y. auto. rewrite subst_open_commute; auto.
  apply_fresh term_arr as y. auto. rewrite* subst_open_commute; auto.
Qed.

Lemma swap_term : forall x z t, 
  x \notin fv t -> z \notin fv t -> 
  term (t^^tm_fv x) -> term (t^^tm_fv z).
intros.
  destruct (x==z).
    subst. auto.

    rewrite (@subst_intro x); auto.
    apply subst_term; auto.
Qed.

Definition P_term_dec (n:nat) := forall t, tm_size t = n ->
  decidable (term t) /\
  decidable (exists v : tm, t = tm_app P_SAYS v /\ term v) /\
  decidable (exists v : tm, t = tm_app PF_SIGN v /\ term v) /\
  decidable (exists v : tm, t = tm_app TM_PF_BIND v /\ term v) /\
  decidable (exists v : tm, t = tm_app PF_SAYS_RETURN v /\ term v) /\
  decidable (exists v : tm, t = tm_app PF_SAYS_BIND v /\ term v) /\
  decidable (exists v1 : tm,
                       exists v2 : tm,
                         exists v3 : tm,
                           t = tm_app (tm_app (tm_app TM_PIF v1) v2) v3 /\ term v1 /\ term v2 /\ term v3)
  .
Lemma _term_dec:  forall n, P_term_dec n.
Proof.
  intro n.
  apply lt_wf_ind. clear n.
  intros n H.
  unfold P_term_dec in *.
  unfold decidable in *.
  destruct t; intros; subst.
  Case "fv".
split. 
   left. apply term_fv.
split;try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; inversion H3].
split;try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; inversion H3].
split;try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; inversion H3].
split;try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; inversion H3].
split;try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; inversion H3].
right; unfold not; intros. inversion H0. inversion H1. inversion H2. decompose [and] H3. inversion H4. 
  Case "bv". 
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; inversion H3].
right; unfold not; intros. inversion H0. inversion H1. inversion H2. decompose [and] H3. inversion H4.  
 Case "con".
split.
  left. apply term_con.
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split;try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; inversion H3].
right; unfold not; intros. inversion H0. inversion H1. inversion H2. decompose [and] H3. inversion H4.  
Case "ctr". 
split.
  left. apply term_ctr.
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split;try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; inversion H3].
right; unfold not; intros. inversion H0. inversion H1. inversion H2. decompose [and] H3. inversion H4.  
  (* app *)
split.
  assert (tm_size t1 < tm_size (tm_app t1 t2)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_app t1 t2)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H1; auto.
  decompose [and] H0. decompose [and] H1.
  clear H0 H1. 
  destruct H4; destruct H8.
  (* SAYS t2*)
  left. inversion H0. inversion H4. subst. apply term_P_SAYS; auto.
  (* ~ t2 *)
  right. unfold not. intro. inversion H4; apply H1; auto.
  (*~SAYS t2*)
  destruct H3.
    (*PF_SIGN*)
    left. inversion H3. inversion H4. subst. apply term_PF_SIGN; auto.
    (*~PF_SIGN*)
    destruct (eq_tm_dec t1 T_PF).
      (*T_PF*)
      subst. left. apply term_T_PF; auto.
      (*~T_PF*)
      destruct (eq_tm_dec t1 TM_PF_RETURN).
        (*TM_PF_RETURN*)
        subst. left. apply term_TM_PF_RETURN; auto.
        (*~TM_PF_RETURN*)          
        destruct H5.
          (*TM_PF_BIND*)
          left. inversion H4. inversion H5. subst. apply term_TM_PF_BIND; auto.
          (*~TM_PF_BIND*)
          destruct (eq_tm_dec t1 TM_SAY).
            (*TM_SAY*)
            subst. left. apply term_TM_SAY; auto.
            (*~TM_SAY*)
            destruct H6.
              (*PF_SAYS_RETURN*)  
              left. inversion H5. inversion H6. subst. apply term_PF_SAYS_RETURN; auto.
              (*~PF_SAYS_RETURN*)  
              destruct H7.
                (*PF_SAYS_BIND*)  
                left. inversion H6. inversion H7. subst. apply term_PF_SAYS_BIND; auto.
                (*~PF_SAYS_BIND*) 
                destruct H9.
                  (*TM_PIF*) 
                  left. inversion H7.  inversion H8.  inversion H9. decompose [and] H15. subst. apply term_PIF; auto.
                  (*~TM_PIF*) 
                  destruct H2.
                    (*t1*)
                    assert (J1:=is_tm_app_dec2 t1 P_SAYS).
                    assert (J2:=is_tm_app_dec2 t1 PF_SIGN).
                    assert (J3:=is_tm_app_dec2 t1 TM_PF_BIND).
                    assert (J4:=is_tm_app_dec2 t1 PF_SAYS_RETURN).
                    assert (J5:=is_tm_app_dec2 t1 PF_SAYS_BIND).
                    assert (J6:=is_tm_app_dec3 t1 TM_PIF).
                    destruct J1.
                      right. unfold not. intros. inversion H9; subst; auto.
                        apply H0. eapply ex_intro; eauto.
                        apply H3. eapply ex_intro; eauto.
                        apply H4. eapply ex_intro; eauto.
                        apply H7. eapply ex_intro; eapply ex_intro; eapply ex_intro; eauto.
                        apply H5. eapply ex_intro; eauto.
                        apply H6. eapply ex_intro; eauto.
                    destruct J2.
                      right. unfold not. intros. inversion H15; subst; auto.
                        apply H0. eapply ex_intro; eauto.
                        apply H3. eapply ex_intro; eauto.
                        apply H4. eapply ex_intro; eauto.
                        apply H7. eapply ex_intro; eapply ex_intro; eapply ex_intro; eauto.
                        apply H5. eapply ex_intro; eauto.
                        apply H6. eapply ex_intro; eauto.
                    destruct J3.
                      right. unfold not. intros. inversion H17; subst; auto.
                        apply H0. eapply ex_intro; eauto.
                        apply H3. eapply ex_intro; eauto.
                        apply H4. eapply ex_intro; eauto.
                        apply H7. eapply ex_intro; eapply ex_intro; eapply ex_intro; eauto.
                        apply H5. eapply ex_intro; eauto.
                        apply H6. eapply ex_intro; eauto.
                    destruct J4.
                      right. unfold not. intros. inversion H18; subst; auto.
                        apply H0. eapply ex_intro; eauto.
                        apply H3. eapply ex_intro; eauto.
                        apply H4. eapply ex_intro; eauto.
                        apply H7. eapply ex_intro; eapply ex_intro; eapply ex_intro; eauto.
                        apply H5. eapply ex_intro; eauto.
                        apply H6. eapply ex_intro; eauto.
                    destruct J5.
                      right. unfold not. intros. inversion H19; subst; auto.
                        apply H0. eapply ex_intro; eauto.
                        apply H3. eapply ex_intro; eauto.
                        apply H4. eapply ex_intro; eauto.
                        apply H7. eapply ex_intro; eapply ex_intro; eapply ex_intro; eauto.
                        apply H5. eapply ex_intro; eauto.
                        apply H6. eapply ex_intro; eauto.
                    destruct J6.
                      right. unfold not. intros. inversion H20; subst; auto.
                        apply H0. eapply ex_intro; eauto.
                        apply H3. eapply ex_intro; eauto.
                        apply H4. eapply ex_intro; eauto.
                        apply H7. eapply ex_intro; eapply ex_intro; eapply ex_intro; eauto.
                        apply H5. eapply ex_intro; eauto.
                        apply H6. eapply ex_intro; eauto.
                    left. apply term_app; auto.
                    (*~t1*)
                    right. unfold not. intro. inversion H8; subst; auto.
                      apply H0. eapply ex_intro; eauto.
                        apply H3. eapply ex_intro; eauto.
                        apply H4. eapply ex_intro; eauto.
                        apply H7. eapply ex_intro; eapply ex_intro; eapply ex_intro; eauto.
                        apply H5. eapply ex_intro; eauto.
                        apply H6. eapply ex_intro; eauto.
  (* ~ t2 *)
  right. unfold not. intro. inversion H4; apply H1; auto.

split.
  destruct t1;
    try solve [right; unfold not; intro; inversion H0; inversion H1; inversion H2].
    unfold P_SAYS.
    destruct (eq_tm_dec (tm_con c) (tm_con 20)).
      assert (tm_size t2 < tm_size (tm_app (tm_con c) t2)). simpl. omega.
      apply H with (t:=t2) in H0; auto.
      destruct H0. destruct H0.      
        inversion e. subst. left. apply ex_intro with (x:=t2); auto.
        right; unfold not; intro. inversion H2. inversion H3.  inversion H4. subst. apply H0; auto.

        right; unfold not; intro. inversion H0. inversion H1.  inversion H2. subst.
        apply n. auto.

split.
  destruct t1;
    try solve [right; unfold not; intro; inversion H0; inversion H1; inversion H2].
    unfold PF_SIGN.
    destruct (eq_tm_dec (tm_con c) (tm_con 52)).
      assert (tm_size t2 < tm_size (tm_app (tm_con c) t2)). simpl. omega.
      apply H with (t:=t2) in H0; auto.
      destruct H0. destruct H0.      
        inversion e. subst. left. apply ex_intro with (x:=t2); auto.
        right; unfold not; intro. inversion H2. inversion H3.  inversion H4. subst. apply H0; auto.

        right; unfold not; intro. inversion H0. inversion H1.  inversion H2. subst.
        apply n. auto.

split.
  destruct t1;
    try solve [right; unfold not; intro; inversion H0; inversion H1; inversion H2].
    unfold TM_PF_BIND.
    destruct (eq_tm_dec (tm_con c) (tm_con 41)).
      assert (tm_size t2 < tm_size (tm_app (tm_con c) t2)). simpl. omega.
      apply H with (t:=t2) in H0; auto.
      destruct H0. destruct H0.      
        inversion e. subst. left. apply ex_intro with (x:=t2); auto.
        right; unfold not; intro. inversion H2. inversion H3.  inversion H4. subst. apply H0; auto.

        right; unfold not; intro. inversion H0. inversion H1.  inversion H2. subst.
        apply n. auto.

split.
  destruct t1;
    try solve [right; unfold not; intro; inversion H0; inversion H1; inversion H2].
    unfold PF_SAYS_RETURN.
    destruct (eq_tm_dec (tm_con c) (tm_con 50)).
      assert (tm_size t2 < tm_size (tm_app (tm_con c) t2)). simpl. omega.
      apply H with (t:=t2) in H0; auto.
      destruct H0. destruct H0.      
        inversion e. subst. left. apply ex_intro with (x:=t2); auto.
        right; unfold not; intro. inversion H2. inversion H3.  inversion H4. subst. apply H0; auto.

        right; unfold not; intro. inversion H0. inversion H1.  inversion H2. subst.
        apply n. auto.

split.
  destruct t1;
    try solve [right; unfold not; intro; inversion H0; inversion H1; inversion H2].
    unfold PF_SAYS_BIND.
    destruct (eq_tm_dec (tm_con c) (tm_con 51)).
      assert (tm_size t2 < tm_size (tm_app (tm_con c) t2)). simpl. omega.
      apply H with (t:=t2) in H0; auto.
      destruct H0. destruct H0.      
        inversion e. subst. left. apply ex_intro with (x:=t2); auto.
        right; unfold not; intro. inversion H2. inversion H3.  inversion H4. subst. apply H0; auto.

      right; unfold not; intro. inversion H0. inversion H1.  inversion H2. subst.
        apply n. auto.

  destruct t1;
    try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; decompose [and] H3; inversion H4].
  destruct t1_1;
    try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; decompose [and] H3; inversion H4].
  destruct t1_1_1;
    try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2; decompose [and] H3; inversion H4].
    unfold TM_PIF.
    destruct (eq_tm_dec (tm_con c) (tm_con 43)).
      assert (tm_size t2 < tm_size (tm_app (tm_app ( tm_app (tm_con c) t1_1_2) t1_2) t2)). simpl. omega.
      assert (tm_size t1_2 < tm_size (tm_app (tm_app ( tm_app (tm_con c) t1_1_2) t1_2) t2)). simpl. omega.
      assert (tm_size t1_1_2 < tm_size (tm_app (tm_app ( tm_app (tm_con c) t1_1_2) t1_2) t2)). simpl. omega.
      apply H with (t:=t2) in H0; auto.
      apply H with (t:=t1_2) in H1; auto.
      apply H with (t:=t1_1_2) in H2; auto.
      decompose [and] H0. decompose [and] H1. decompose [and] H2. 
      destruct H3; destruct H9; destruct H16.
        inversion e. unfold TM_PIF in *. subst. left.
          apply ex_intro with (x:=t1_1_2). 
          apply ex_intro with (x:=t1_2). 
          apply ex_intro with (x:=t2). 
          auto.
        right; unfold not; intro. inversion H23. inversion H25.  inversion H26. 
          decompose [and] H27. inversion H28; subst. apply H16; auto.
        right; unfold not; intro. inversion H23. inversion H25.  inversion H26. 
          decompose [and] H27. inversion H28; subst. apply H9; auto.
        right; unfold not; intro. inversion H23. inversion H25.  inversion H26. 
          decompose [and] H27. inversion H28; subst. apply H16; auto.
        right; unfold not; intro. inversion H23. inversion H25.  inversion H26. 
          decompose [and] H27. inversion H28; subst. apply H3; auto.
        right; unfold not; intro. inversion H23. inversion H25.  inversion H26. 
          decompose [and] H27. inversion H28; subst. apply H16; auto.
        right; unfold not; intro. inversion H23. inversion H25.  inversion H26. 
          decompose [and] H27. inversion H28; subst. apply H9; auto.
        right; unfold not; intro. inversion H23. inversion H25.  inversion H26. 
          decompose [and] H27. inversion H28; subst. apply H3; auto.

     right; unfold not; intro. inversion H0. inversion H1.  inversion H2.
        decompose [and] H3. inversion H4. subst.
        apply n. auto.

  (* abs *)
split.
  assert (tm_size t1 < tm_size (tm_abs t1 t2)). simpl. omega.
  pick_fresh y.
  assert (tm_size (t2^^(tm_fv y)) < tm_size (tm_abs t1 t2)).
    simpl. unfold open_tm. rewrite open_tm_size_eq. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2^^(tm_fv y)) in H1; auto.
  destruct H0; destruct H1.
  destruct H0; destruct H1.
    left. apply term_abs with (L:=fv t1 \u fv t2); auto.
    intros. apply swap_term with (x:=y); auto.

    right. unfold not. intro. inversion H4.
    pick_fresh x. 
    assert (x \notin L). auto. 
    apply H8 in H9.
    apply H1. apply swap_term with (x:=x); auto.

    right. unfold not. intro. inversion H4. contradiction.
    right. unfold not. intro. inversion H4. contradiction.

split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].   (* arr *)
right; unfold not; intros. inversion H0. inversion H1. inversion H2. decompose [and] H3. inversion H4.  

  (* arr *)
split.
  assert (tm_size t1 < tm_size (tm_abs t1 t2)). simpl. omega.
  pick_fresh y.
  assert (tm_size (t2^^(tm_fv y)) < tm_size (tm_abs t1 t2)).
    simpl. unfold open_tm. rewrite open_tm_size_eq. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2^^(tm_fv y)) in H1; auto.
  destruct H0; destruct H1.
  destruct H0; destruct H1.
    left. apply term_arr with (L:=fv t1 \u fv t2); auto.
    intros. apply swap_term with (x:=y); auto.

    right. unfold not. intro. inversion H4.
    pick_fresh x. 
    assert (x \notin L). auto. 
    apply H8 in H9.
    apply H1. apply swap_term with (x:=x); auto.

    right. unfold not. intro. inversion H4. contradiction.
    right. unfold not. intro. inversion H4. contradiction.
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].   (* arr *)
right; unfold not; intros. inversion H0. inversion H1. inversion H2. decompose [and] H3. inversion H4.  

  (* mat *)
split.
  assert (tm_size t1 < tm_size (tm_mat t1 t2 t3)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_mat t1 t2 t3)). simpl. omega.
  assert (tm_size t3 < tm_size (tm_mat t1 t2 t3)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H1; auto.
  apply H with (t:=t3) in H2; auto.
  destruct H0; destruct H1; destruct H2.
  destruct H0; destruct H1; destruct H2;
    try solve [right; unfold not; intro; inversion H6;contradiction].
    left. apply term_mat; auto.
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].   (* arr *)
right; unfold not; intros. inversion H0. inversion H1. inversion H2. decompose [and] H3. inversion H4.  

  (* brn *)
split.
  assert (tm_size t1 < tm_size (tm_brn c t1 t2)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_brn c t1 t2)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H1; auto.
  destruct H0; destruct H1.
  destruct H0; destruct H1;
    try solve [right; unfold not; intro; inversion H4;contradiction].
    left. apply term_brn; auto.
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].   (* arr *)
right; unfold not; intros. inversion H0. inversion H1. inversion H2. decompose [and] H3. inversion H4.  

  (* cast *)
split.
  assert (tm_size t1 < tm_size (tm_cast t1 t2)). simpl. omega.
  assert (tm_size t2 < tm_size (tm_cast t1 t2)). simpl. omega.
  apply H with (t:=t1) in H0; auto.
  apply H with (t:=t2) in H1; auto.
  destruct H0; destruct H1.
  destruct H0; destruct H1;
    try solve [right; unfold not; intro; inversion H4;contradiction].
    left. apply term_cast; auto.
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].
split; try solve [right; unfold not; intros; inversion H0; inversion H1; inversion H2].   (* arr *)
right; unfold not; intros. inversion H0. inversion H1. inversion H2. decompose [and] H3. inversion H4.  
Qed.

Lemma term_dec: forall t, decidable (term t).
intros.
  assert (P_term_dec (tm_size t)).
    apply (@_term_dec (tm_size t)).
  unfold P_term_dec in H.
  assert (tm_size t = tm_size t). auto.
  decompose [and] (@H t H0).
  auto.
Qed.

Lemma wf_tm_gives_term_mutual: 
forall S1 S2, 
  (forall E e t, wf_tm S1 S2 E e t -> term e)
/\
  (forall E s  args b  t_res,
    wf_branches  S1 S2 E s args b t_res -> term b).
Proof.
  intros.
  apply mutual_wf_tm_branches_ind; eauto; intros.
    unfold K_TYPE. apply term_con.
    unfold K_PROP. apply term_con.
    (*app*)    
    assert (~exists v : tm, e1 = tm_app P_SAYS v) as J1.
      unfold not. intro. inversion H1. subst.
      inversion w. subst. inversion H6. 
    assert (~exists v : tm, e1 = tm_app PF_SIGN v) as J2.
      unfold not. intro. inversion H1. subst.
      inversion w. subst. inversion H6. 
    assert (~exists v : tm, e1 = tm_app TM_PF_BIND v) as J3.
      unfold not. intro. inversion H1. subst.
      inversion w. subst. inversion H6. 
    assert (~exists v : tm, e1 = tm_app PF_SAYS_RETURN v) as J4.
      unfold not. intro. inversion H1. subst.
      inversion w. subst. inversion H6. 
    assert (~exists v : tm, e1 = tm_app PF_SAYS_BIND v) as J5.
      unfold not. intro. inversion H1. subst.
      inversion w. subst. inversion H6. 
    assert (~exists v1 : tm,
                       exists v2 : tm,
                         exists v3 : tm,
                           e1 = tm_app (tm_app (tm_app TM_PIF v1) v2) v3) as J6.
      unfold not. intro. inversion H1; subst.  inversion H2; subst.  inversion H3; subst. 
      inversion w; subst. 
      inversion H8; subst. 
      inversion H12; subst. 
      inversion H16. 
    assert (e1 <> T_PF) as J7.
      unfold not. intro. subst. inversion w. 
    assert (e1 <> TM_PF_RETURN ) as J8.
      unfold not. intro. subst. inversion w. 
    assert (e1 <> TM_SAY ) as J9.
      unfold not. intro. subst. inversion w. 
   apply term_app; auto.
   (*prin*)
   unfold T_PRIN. apply term_con.
   (*PIF*)
   pick_fresh x. assert (x \notin L). auto. 
   apply term_PIF; auto.
  apply (H1 x  H4). 

   unfold TM_SELF. apply term_con.
   unfold TM_NONE. apply term_con.
Qed.

Lemma wf_tm_gives_term: forall S1 S2 E t k, 
 wf_tm S1 S2 E t k -> term t.
Proof. 
 intros.
 destruct (wf_tm_gives_term_mutual S1 S2).
 apply H0 in H; auto.
Qed.

Lemma constructs_dec: forall t k, decidable (constructs t k).
Proof. 
  intros. unfold decidable.
  gen k.
  induction t; intros;
    try solve [
      destruct k; 
      try solve [
         right; unfold not; intros; inversion H
        |destruct (eq_tm_dec t1 k1); destruct (eq_tm_dec t2 k2);
          try solve [
             right; unfold not; intros; inversion H; subst; apply n; auto
            |subst; left; apply constructs_base; auto; unfold not; intro; inversion H]
      ]
    ].
    destruct k; try solve [right; unfold not; intros; inversion H].
      destruct (a==a0). 
        subst. left. apply constructs_base; auto. unfold not. intro. inversion H.
        right; unfold not; intros; inversion H. subst. apply n; auto.
    destruct k; try solve [right; unfold not; intros; inversion H].
      destruct (eq_tm_dec (tm_bv b) (tm_bv b0)).
        inversion e. subst. left. apply constructs_base; auto. unfold not. intro. inversion H.
        right; unfold not; intros; inversion H. subst. apply n; auto.
    destruct k; try solve [right; unfold not; intros; inversion H].
      destruct (eq_tm_dec (tm_con c) (tm_con c0)).
        inversion e. subst. left. apply constructs_base; auto. unfold not. intro. inversion H.
        right; unfold not; intros; inversion H. subst. apply n; auto.
    destruct k; try solve [right; unfold not; intros; inversion H].
      destruct (eq_tm_dec (tm_ctr c) (tm_ctr c0)).
        inversion e. subst. left. apply constructs_base; auto. unfold not. intro. inversion H.
        right; unfold not; intros; inversion H. subst. apply n; auto.
    destruct (@IHt2 k).
      left. apply constructs_arr;auto.
      right; unfold not; intros. inversion H0.
        apply H1. apply is_tm_arr_arr.
        apply H; auto.
    destruct k; try solve [right; unfold not; intros; inversion H].
      destruct (eq_tm_dec t1 k1); destruct (eq_tm_dec t2 k2); destruct (eq_tm_dec t3 k3);
        try solve [right; unfold not; intros; inversion H; subst; apply n; auto].
        subst. left. apply constructs_base; auto. unfold not. intro. inversion H.
    destruct k; try solve [right; unfold not; intros; inversion H].
      destruct (eq_tm_dec (tm_ctr c) (tm_ctr c0)); destruct (eq_tm_dec t1 k1); destruct (eq_tm_dec t2 k2);
        try solve [right; unfold not; intros; inversion H; subst; apply n; auto].
        inversion e. subst. left. apply constructs_base; auto. unfold not. intro. inversion H.
Qed.

Lemma constructs_dec2: forall t, decidable (exists k, constructs t k).
Proof. 
  intros. unfold decidable.
  induction t; intros.
    left. apply ex_intro with (x:=tm_fv a). apply constructs_base; auto.
      unfold not. intro. inversion H.
    left. apply ex_intro with (x:=tm_bv b). apply constructs_base; auto.
      unfold not. intro. inversion H.
    left. apply ex_intro with (x:=tm_con c). apply constructs_base; auto.
      unfold not. intro. inversion H.
    left. apply ex_intro with (x:=tm_ctr c). apply constructs_base; auto.
      unfold not. intro. inversion H.
    left. apply ex_intro with (x:=tm_app t1 t2). apply constructs_base; auto.
      unfold not. intro. inversion H.
    left. apply ex_intro with (x:=tm_abs t1 t2). apply constructs_base; auto.
      unfold not. intro. inversion H.

    destruct (@IHt2).
      destruct H. left. apply ex_intro with (x:=x); auto.
      right. unfold not; intros. inversion H0. inversion H1.
        apply H2. apply is_tm_arr_arr.
        apply H. apply ex_intro with (x:=x); auto.

    left. apply ex_intro with (x:=tm_mat t1 t2 t3). apply constructs_base; auto.
      unfold not. intro. inversion H.
    left. apply ex_intro with (x:=tm_brn c t1 t2). apply constructs_base; auto.
      unfold not. intro. inversion H.
    left. apply ex_intro with (x:=tm_cast t1 t2). apply constructs_base; auto.
      unfold not. intro. inversion H.
Qed.

Lemma wf_bundle_tctrs_dec: forall S d k,
 wf_sig S ->
 wf_env S S empty ->
 decidable (wf_bundle_tctrs S d k).
Proof. 
  intros. unfold decidable.
  gen S k.
  induction d; intros.
    destruct (term_dec k).
      (* term k*)
      assert (Jext: sig_extend S S). unfold sig_extend; auto.
      destruct (type_checking_dec H H0 Jext H1); auto. destruct H2.
        (* wf_tm S empty k t*)
        destruct H2 as [t H2].
        destruct (eq_tm_dec S_KIND t).
          subst. left. apply wf_bundle_tctrs_nil; auto.
          right. unfold not. intros. inversion H4.
            assert (t = S_KIND). eapply wf_tm_unique; eauto.
            subst. apply n; auto.
        (* ~wf_tm S empty k t*)
        right. unfold not. intros. inversion H4.
        apply H2. eapply ex_intro; eauto.
      (* ~term k*)
      right. unfold not. intros. inversion H2.
      apply wf_tm_gives_term in H3. apply H1; auto.

   destruct a. destruct p.
   destruct (constructs_dec t k).
     destruct (term_dec t).
      assert (Jext: sig_extend S S). unfold sig_extend; auto.
       destruct (type_checking_dec H H0 Jext H2); auto.
       destruct H3.
         destruct H3 as [tt H3].
         destruct (eq_tm_dec S_KIND tt).
           destruct (IHd S H H0 k).
              subst. destruct (In_dec eq_ctr_dec c0 (dom_defns d)).
                right. unfold not. intros. inversion H6; subst. apply H13; auto.
                destruct (In_dec eq_ctr_dec c0 (dom_ctr_decls c)).
                  right. unfold not. intros. inversion H6; subst. apply H15; auto.
                  left. apply wf_bundle_tctrs_cons; auto.
              right. unfold not. intros. inversion H6; subst. apply H5; auto.
           right. unfold not. intros. inversion H5.
           assert (tt = S_KIND). eapply wf_tm_unique; eauto.
             subst. apply n; auto.

         right. unfold not. intros. inversion H5; subst.
         apply H3. eapply ex_intro; eauto.
          
       right. unfold not. intros. inversion H3; subst.
       apply wf_tm_gives_term in H13. apply H2; eauto.

     right. unfold not. intros. inversion H2. apply H1; auto.
Qed.


Lemma disjoint_dec:  forall l1 l2,  decidable (disjoint l1 l2).
Proof.
introv. 
  induction l1.
  left; unfold disjoint; intros; auto.
  destruct (In_dec eq_ctr_dec a l2).
  right; intro J; unfold disjoint in J; apply (J a); auto.
  destruct IHl1; auto.
  left; unfold disjoint; unfold not; intros; auto.
  destruct (in_inv H0); subst; auto.
  unfold disjoint in H; apply (H n0); auto.
  right; intro; apply H; auto.
  unfold disjoint; unfold not; intros.
  unfold disjoint in H0. apply (H0 n0); auto.
Qed.

Lemma has_level_dec: forall t,
  decidable(exists m, has_level t m).
intros. unfold decidable.
  induction t;
    try solve [left; apply ex_intro with (x:=0); apply has_level_base; unfold not; intros; inversion H].
    destruct (@IHt2).
         inversion H. left. apply ex_intro with (x:=1+x). apply has_level_arr; auto.  
         right; unfold not; intros. inversion H0; auto. apply H. inversion H1.
           assert (False). apply H2; auto. inversion H5.
           apply ex_intro with (x:=n); auto.
Qed.

Lemma all_tm_bv_dec: forall n l,
  decidable(all_tm_bv n l).
intros. unfold decidable.
  gen n.
  induction l; intros.
    left. apply all_tm_bv_nil.
    destruct (@IHl (n+1)).
      destruct a;  
         try solve [right; unfold not; intros; inversion H0; auto].
         destruct (n===b).
            left. subst. apply all_tm_bv_cons; auto.  
            right; unfold not; intros. inversion H0; auto.
      right; unfold not; intros. inversion H0; auto.
Qed.

Lemma Core_positive_dec : forall (P: list ctr) t, 
 decidable (Core.positive P t).
intros. unfold decidable.
  gen P. induction t; intros;
     try solve[right; unfold not; intros; inversion H; subst].
     left. apply postive_ctr.

     destruct (disjoint_dec P (ctrs_of t2)).
     destruct (@IHt1 P).
       left. apply positive_app; auto.
       right. unfold not. intro. inversion H1. apply H0; auto.
     right. unfold not. intro. inversion H0. apply H; auto.

     destruct (disjoint_dec P (ctrs_of t1)).
     destruct (@IHt2 P).
       left. apply positive_arr; auto.
       right. unfold not. intro. inversion H1. apply H0; auto.
     right. unfold not. intro. inversion H0. apply H; auto.
Qed.

Lemma wf_ctr_decls_dec: forall P S1 S2 c t cdl, 
 wf_sig S1 ->
 wf_env S1 S2 empty ->
 sig_extend S1 S2 ->
 decidable (wf_ctr_decls P S1 S2 c t cdl).
Proof.
intros.  unfold decidable.
  gen P S1 S2 c t.
  induction cdl; intros.
    left. apply wf_ctr_decls_nil.

   destruct a. destruct c0.
   destruct (term_dec t0).
     (*term t0*)
     destruct (type_checking_dec  H H0 H1 H2); auto. destruct H3.
       (*wf_tm t0*)
       destruct H3 as [tt H3].
          destruct (IHcdl P S1  H S2 H0 H1 c t).
            (*wf_ctr_decls*)
            destruct (In_dec eq_ctr_dec (n, n0) (dom_ctr_decls cdl)).
                (*~ In (ctr, n) (dom_ctr_decls cdl)*)
                right. unfold not. intros. inversion H6; subst. contradiction.
                (* In (ctr, n) (dom_ctr_decls cdl)*)
                destruct (constructs_dec2 t0).
                   (*construct*)
                   destruct H6 as [c_inst H6].
                   destruct(fully_applied_typ_dec c_inst c t).
                     (*full_applied_typ*)
                     destruct(fully_applied_args_dec2 c_inst c).
                       (*full_applied_arg*)
                       destruct H8 as [args H8].
                       destruct (n0 === (length args)).
                         (*n0 = (length args)*)
                         destruct (has_level_dec t0).
                           (*has_level*)
                           destruct H9 as [m H9].
                           destruct (le_lt_dec n0 m).
                              (*m<=n0*)
                              destruct (all_tm_bv_dec (m-n0) args).
                                 (*all_tm_bv_dec*)
                                 destruct (Core_positive_dec P t0).
                                   (*positive*)
                                   left. eapply wf_ctr_decls_cons; eauto.
                                     omega.
                                   (*~positive*)
                                  right. unfold not. intros. inversion H12; subst.
                                  assert (c_inst = c_inst0). eapply construct_unique; eauto. 
                                  subst.
                                  assert (args0 = args).
                                    destruct (@fully_applied_args_unique c_inst0 c args0 c args); auto.
                                  subst.
                                  assert (m = m0).
                                     apply (@has_level_unique t0 m m0); auto.
                                  subst. contradiction.
                                 (*~all_tm_bv_dec*)
                                right. unfold not. intros. inversion H11; subst.
                                assert (c_inst = c_inst0). eapply construct_unique; eauto. 
                                subst.
                                assert (args0 = args).
                                  destruct (@fully_applied_args_unique c_inst0 c args0 c args); auto.
                                subst.
                                assert (m = m0).
                                   apply (@has_level_unique t0 m m0); auto.
                                subst. contradiction.
                              (*m>n0*)
                              right. unfold not. intros. inversion H10; subst.
                              assert (c_inst = c_inst0). eapply construct_unique; eauto. 
                              subst.
                              assert (args0 = args).
                                destruct (@fully_applied_args_unique c_inst0 c args0 c args); auto.
                              subst.
                              assert (m = m0).
                                 apply (@has_level_unique t0 m m0); auto. 
                              subst. contradiction.
                           (*~has_level*)
                           right. unfold not. intros. inversion H10; subst. apply H9.  apply ex_intro with (x:=m); auto.
                         (*~n0 = (length args)*)
                         right. unfold not. intros. inversion H9; subst.
                         assert (c_inst = c_inst0). eapply construct_unique; eauto. 
                         subst.
                         assert (args0 = args).
                           destruct (@fully_applied_args_unique c_inst0 c args0 c args); auto.
                         subst. apply n2; auto.
                       (*~full_applied_arg*)
                       right. unfold not. intros. inversion H9; subst.
                       apply H8. apply ex_intro with (x:=args). auto.
                       assert (c_inst = c_inst0). eapply construct_unique; eauto. 
                       subst.  auto.
                     (*~full_applied_typ*)
                     right. unfold not. intros. inversion H8; subst. 
                     assert (c_inst = c_inst0). eapply construct_unique; eauto. 
                     subst. contradiction.
                   (*~construct*)
                  right. unfold not. intros. inversion H7; subst. apply H6. apply ex_intro with (x:=c_inst); auto.
            (*~wf_ctr_decls*)
           right. unfold not. intros. inversion H6; subst. contradiction.
         (*~wf_tm t0*)
         right. unfold not. intros. inversion H5; subst.
         apply H3. eapply ex_intro; eauto.
       (*~term t0*)
       right. unfold not. intros. inversion H3; subst.
       apply wf_tm_gives_term in H10. contradiction.
Qed.

Lemma wf_bundle_ctrs_dec: forall P S1 S2 d, 
 wf_sig S1 ->
 wf_env S1 S2 empty ->
 sig_extend S1 S2 ->
 decidable (wf_bundle_ctrs P S1 S2 d).
Proof.
intros. unfold decidable.
  gen P S1 S2.
  induction d; intros.
    left. apply wf_bundle_ctrs_nil.

    destruct a. destruct p. 
    destruct (@IHd P S1 H S2); auto.
      destruct (@wf_ctr_decls_dec P S1  S2 c0 t c); auto.
        destruct (disjoint_dec (dom_ctr_decls c) (dom_defns d)).
          left. apply wf_bundle_ctrs_cons; auto.
          right. unfold not. intros. inversion H5; subst. contradiction.
        right. unfold not. intros. inversion H4; subst. contradiction.
      right. unfold not. intros. inversion H3; subst. contradiction.
Qed.

Lemma positive_weakening: forall (P: list ctr) t, 
 Core.positive P t -> Core.positive nil t.
Proof.
 introv Hp.
 induction Hp; subst; auto.
 eapply positive_arr; eauto.
 unfold disjoint; intro; auto.
 eapply positive_app; eauto.
 unfold disjoint; intro; auto.
Qed.


Lemma wf_ctr_decls_weakening:  forall P S1 S2 c k d,
 wf_ctr_decls P S1 S2 c k d ->
 wf_ctr_decls nil S1 S2 c k d.
Proof.
 introv Hwf.
 induction Hwf; subst; auto.
 eapply wf_ctr_decls_cons; eauto.
 eapply positive_weakening; eauto.
Qed.

Lemma wf_bundle_ctrs_weakening:  forall P S1 S2 d,
 wf_bundle_ctrs P S1 S2 d ->
 wf_bundle_ctrs nil S1 S2 d.
Proof.
 introv Hwf.
 induction Hwf; subst; auto.
 eapply wf_bundle_ctrs_cons; eauto.
 eapply wf_ctr_decls_weakening; eauto.
Qed.

Lemma wf_tctrs_gives_wf_ctrs:  forall P S1 S2 d K,
  wf_bundle_tctrs S1 d K ->
  wf_bundle_ctrs P S1 S2 (get_tctr_defns_of d).
Proof.
 introv Hwf_tctrs.
 induction d; auto.
 destruct a. destruct p.
 simpl.
 inversion Hwf_tctrs; subst.
 apply wf_bundle_ctrs_cons; auto.
 simpl. unfold disjoint; unfold not; intros. inversion H.
Qed.

Lemma wf_bundle_tctrs_weakening: forall S d K a,
  wf_sig S -> disjoint (dom_sig(a::nil)) (dom_sig S) ->
   wf_bundle_tctrs S d K -> wf_bundle_tctrs (a::S) d K.
Proof.
 introv Hwf_sig Hdisj Hwf_tctrs.
 induction d; auto; inversion Hwf_tctrs; subst; constructor; auto.
replace (a::S) with (a::nil++S); auto. 
eapply wf_tm_weakening_sig2_r with (S:=a::nil) (S1:=S) (S2:=S); eauto.
unfold sig_extend; auto.
replace (a::S) with (a::nil++S); auto. 
eapply wf_tm_weakening_sig2_r with (S:=a::nil) (S1:=S) (S2:=S); eauto.
unfold sig_extend; auto.
Qed.

Lemma wf_tctrs_inv : forall S d K, 
  wf_bundle_tctrs S d K -> wf_bundle_tctrs S (get_tctr_defns_of d) K .
Proof. 
 introv Htctrs.
 induction d; auto.
 inversion Htctrs; subst.
 simpl. 
 apply wf_bundle_tctrs_cons; auto.
 unfold not; intro.  unfold not in H2; apply H2. 
 eapply in_tctr_defns_in; eauto. 
Qed.

Lemma wf_tctrs_gives_wf_sig: forall S d K,
  wf_sig S -> 
  wf_bundle_tctrs S d K ->
  K = K_TYPE \/ K = K_PROP ->
  disjoint (dom_defns d) (dom_sig S) -> 
  wf_sig (bundle_defns(get_tctr_defns_of d)::S).
Proof.
 introv Hwf_sig Hwf_tctrs Hk Hdisj.
 destruct Hk.
 Case "K_TYPE".
 apply wf_sig_cons_Type; auto. subst. 
  apply wf_tctrs_inv ; auto.
  eapply  defns_disjoint_weakening; eauto.
  eapply wf_tctrs_gives_wf_ctrs; eauto.
  eapply wf_bundle_tctrs_weakening; eauto.
  unfold dom_sig; simpl.  rewrite <- app_nil_end.
  replace (flat_map dom_bundle S) with (dom_sig S); auto. eapply defns_disjoint_weakening; eauto.
  eapply defns_disjoint_weakening; eauto.
 Case "K_PROP".
 apply wf_sig_cons_Prop; auto. subst. 
  apply wf_tctrs_inv ; auto.
  eapply  defns_disjoint_weakening; eauto.
  eapply wf_tctrs_gives_wf_ctrs; eauto.
  eapply wf_bundle_tctrs_weakening; eauto.
  unfold dom_sig; simpl.  rewrite <- app_nil_end.
  replace (flat_map dom_bundle S) with (dom_sig S); auto. eapply defns_disjoint_weakening; eauto.
  eapply defns_disjoint_weakening; eauto.
Qed.

Lemma wf_sig_dec: forall S, decidable (wf_sig S).
Proof. 
introv.
induction S.
Case "nil".
 left; auto.
Case "cons".
  destruct IHS.
  Case "wf_sig S". 
    assert (Jwf_env: wf_env S S empty); auto.
    destruct a.
  Case "bundle_defns".
    destruct (wf_bundle_tctrs_dec d K_TYPE H Jwf_env).
    assert (J:= disjoint_dec (dom_defns d) (dom_sig  S)). 
    destruct J.
     assert (Jwf_sig: wf_sig ((bundle_defns (get_tctr_defns_of d))::S)).
     eapply wf_tctrs_gives_wf_sig; eauto.
     assert (Jwf_env': wf_env  ((bundle_defns (get_tctr_defns_of d))::S) (bundle_defns d::S)  empty); auto.
    destruct (wf_bundle_ctrs_dec nil d  Jwf_sig Jwf_env') .
    unfold sig_extend; right; eauto.
    left; apply wf_sig_cons_Type; eauto.
    Case "~wf_bundle_ctrs".
    right; intro J; inversion J; subst; apply H2. contradiction.
    eapply wf_bundle_ctrs_weakening; eauto.
    Case "~disjoing (dom_defns d) (dom_sig S))".
    right; intro J; inversion J; subst; contradiction.
    Case "~wf_bundle_tctrs S d K_TYPE".
    destruct (wf_bundle_tctrs_dec d K_PROP H Jwf_env).
    destruct (disjoint_dec (dom_defns d) (dom_sig  S)).
     assert (Jwf_sig: wf_sig ((bundle_defns (get_tctr_defns_of d))::S)).  
     eapply wf_tctrs_gives_wf_sig; eauto.
     assert (Jwf_env': wf_env  ((bundle_defns (get_tctr_defns_of d))::S) (bundle_defns d::S)  empty); auto.
    destruct (wf_bundle_ctrs_dec (dom_defns d) d  Jwf_sig Jwf_env') . 
    unfold sig_extend; right; eauto.
    left; apply wf_sig_cons_Prop; eauto.
    Case "~wf_bundle_ctrs".
    right; intro J; inversion J; subst; contradiction.
    Case "~disjoing (dom_defns d) (dom_sig S))".
    right; intro J; inversion J; subst; contradiction.
    Case "~wf_bundle_tctrs S d K_TYPE \/ K_PROP".
    right; intro J; inversion J; subst; contradiction.
  Case "bundle_assert".
    destruct (term_dec t).
    assert (Jext: sig_extend S S). unfold sig_extend; auto.
    destruct (type_checking_dec  H Jwf_env Jext H0).  clear H2.
    destruct H1. destruct H1 as [k Jwft].
    destruct (eq_tm_dec k S_KIND).
    destruct (constructs_dec t K_PROP). 
    destruct (In_dec eq_ctr_dec c (dom_sig S)).
    Case "In c (dom_sig S)".
    right; intro J; inversion J; subst; contradiction.
    Case "~In c (dom_sig S)".
    subst; left; eapply wf_assertion; eauto.
    Case "~constructs t K_PROP".
    subst; right; intro J; inversion J; subst; contradiction.
    Case "k<>S_KIND".
     subst; right; intro J; inversion J; subst; apply n. 
     eapply wf_tm_unique with (S1:=S) (S2:=S); eauto.
    Case "~t:k".
     subst; right; intro J; inversion J; subst; apply H1;  eauto.
    Case "~term t".
     subst; right; intro J; inversion J; subst; apply H0;  eauto.
     eapply wf_tm_gives_term; eauto.
  Case "~wf_sig S".
   right; intro J; inversion J; subst; apply H; auto.
Qed.
 
Lemma ctr_decl_dec: forall (x y:(ctr*tm)),
  {x = y} + {x <> y}.
intros.
  destruct x; destruct y.
  destruct (eq_ctr_dec c c0).
    destruct (eq_tm_dec t t0).
      subst. left. reflexivity.
      right. unfold not. intros. inversion H. contradiction.
    right. unfold not. intros. inversion H. contradiction.
Qed.

Lemma data_decl_dec: forall d1 d2:data_decl,
  {d1 = d2} + {d1 <> d2}.
intros.
  destruct d1; destruct d2. destruct p; destruct p0.
  destruct (list_eq_dec ctr_decl_dec c c0).
  destruct (eq_ctr_dec c1 c2).
  destruct (eq_tm_dec t t0).
    subst. left. reflexivity.
  right. unfold not. intros. inversion H. contradiction.
  right. unfold not. intros. inversion H. contradiction.
  right. unfold not. intros. inversion H. contradiction.
Qed.   

Lemma bundle_dec: forall b1 b2:bundle,
  {b1 = b2} + {b1 <> b2}.
intros.
  destruct b1; destruct b2.
    destruct d; destruct d0.
      left. reflexivity.
      right. unfold not. intros. inversion H.
      right. unfold not. intros. inversion H.

      destruct (list_eq_dec data_decl_dec d1 d2).
      destruct (data_decl_dec d d0).
        subst. left. reflexivity.
      right. unfold not. intros. inversion H. contradiction.
      right. unfold not. intros. inversion H. contradiction.

    right. unfold not. intros. inversion H.
    right. unfold not. intros. inversion H.

    destruct (eq_ctr_dec c c0).
    destruct (eq_tm_dec t t0).
      subst. left. reflexivity.
    right. unfold not. intros. inversion H. contradiction.
    right. unfold not. intros. inversion H. contradiction.
Qed.

Lemma _sig_extend_dec': forall b b0 S1,
  decidable (exists d, exists S, b::S1 = bundle_defns(get_tctr_defns_of d)::S 
                   /\ b0::S1 = bundle_defns d::S /\ disjoint(dom_defns d) (dom_sig S)).
Proof.
 intros.
  destruct b0. destruct b. 
   gen d0.
   induction d; intros.
   destruct d0.
   left; exists (nil:defns) S1; simpl; repeat split; eauto.
     unfold disjoint. intros. inversion H.

    right; intro J. firstorder. inversion H0; subst; auto. inversion H.

    destruct d0. 
    right; intro J.  destruct J as [d0 [S [J1 [J2 J3]]]].
        inversion J2; subst. inversion J1.

    destruct d0. destruct p. destruct a. destruct p.
    destruct (list_eq_dec ctr_decl_dec c nil). 
    destruct (eq_ctr_dec c0 c2). 
    destruct (eq_tm_dec t t0).
    destruct (In_dec (eq_ctr_dec) c2 (dom_sig S1)).
    
    Case "In c2 S1".
     right; intro J. destruct J as [d0 [S [J1 [J2 J3]]]].
     inversion J2; subst.  unfold dom_defns in J3. 
     unfold disjoint in J3. simpl in J3.  destruct (J3 c2); auto.
     
    destruct (disjoint_dec (dom_ctr_decls c1) (dom_sig S1)).
    subst.
    destruct (IHd d1); auto.
    destruct H0 as  [d0 [S [J1 [J2 J3]]]].
    inversion J1; subst. inversion J2; subst.
    left; exists ((c2,t0,c1)::d0) S; repeat split; eauto.
    unfold disjoint; unfold not; intros.
    unfold dom_defns in H0. simpl in H0.
    destruct H0; subst; auto.
    fold dom_defns in H0.
    destruct (in_app_or  (dom_ctr_decls c1) (dom_defns d0) n0 H0).
    destruct (H n0 H2 H1).
    destruct (J3 n0 H2 H1).
    
    Case "~IH".
    right;  intro J; apply H0; destruct J as [d0 [S [J1 [J2 J3]]]]. clear IHd H0.
     inversion J2. destruct d0; auto. inversion H1.
     inversion H1.   simpl in J1. inversion J1. 
     exists d2 S; subst; repeat split; auto. 
     
     unfold disjoint; unfold not; intros.
     assert (Jin: In n0 (dom_defns ((c2, t0,c1)::d2))).
     unfold dom_defns; simpl; right; fold dom_defns; eapply in_or_app; eauto.
     destruct (J3 n0 Jin H2).
     
     right;  intro J; apply H; destruct J as [d0 [S [J1 [J2 J3]]]]. clear IHd H.
     inversion J2; subst. 
     unfold disjoint; unfold not; intros.
     assert (Jin: In n0 (dom_defns ((c2, t0,c1)::d))).
     unfold dom_defns; simpl; right; fold dom_defns; eapply in_or_app; eauto.
     destruct (J3 n0 Jin H0).
     
     right; intro J. destruct J as [d0 [S [J1 [J2 J3]]]].
     inversion J2; subst. simpl in J1. inversion J1; subst. apply n; auto.
    
    right; intro J. destruct J as [d0 [S [J1 [J2 J3]]]].
     inversion J2; subst. simpl in J1. inversion J1; subst. apply n; auto.
 
    right; intro J. destruct J as [d0 [S [J1 [J2 J3]]]].
     inversion J2; subst. simpl in J1. inversion J1; subst. apply n; auto. 

    right; intro J. firstorder. inversion H.

  right; intro J. firstorder. inversion H0.
Qed.


Lemma _sig_extend_dec: forall S1 S2,
   decidable 
     (exists d : defns,
       exists S : list bundle,
         S1 = bundle_defns (get_tctr_defns_of d) :: S /\
         S2 = bundle_defns d :: S /\ disjoint (dom_defns d) (dom_sig S)).
destruct S1; destruct S2; intros;
  try solve [right; unfold not; intro; inversion H; inversion H0; decompose [and] H1; inversion H2].
  right. unfold not. intro. inversion H. inversion H0. decompose [and] H1. inversion H4.
  destruct (list_eq_dec bundle_dec S1 S2).

  subst; eapply _sig_extend_dec'; eauto. 

    right. unfold not. intro. inversion H. inversion H0. decompose [and] H1. clear H H1 H0.
      inversion H2. inversion H4. subst. apply n; auto.
Qed. 

Lemma sig_extend_dec: forall S1 S2,
  decidable (sig_extend S1 S2).
intros.
  unfold sig_extend.
  destruct (list_eq_dec bundle_dec S1 S2).
    left. left. auto.
  destruct (_sig_extend_dec S1 S2).
    left. right. auto.
  right. unfold not. intro. destruct H0; auto.
Qed.

Lemma wf_env_dec: forall S1 S2 E,
 wf_sig S1 -> ok E -> 
 sig_extend S1 S2 ->
 decidable (wf_env S1 S2 E).
Proof.
introv Hwf_sig HokE Hsigex.
gen S1 S2.
induction HokE; intros.
left; auto.
destruct (@IHHokE S1 Hwf_sig S2 Hsigex).
 (*wf_env*)
 destruct a.
 Case "x~:t".
   destruct (term_dec t).
     (*term t*)
       destruct (type_checking_dec Hwf_sig H0 Hsigex H1 ); auto. 
       destruct H2.
         (*wf_tm t0*)
         left. destruct H2. eapply wf_env_cons_var; eauto.
         (*~wf_tm t0*)
         right; intro J; inversion J; subst. apply H2. apply ex_intro with (x:=k); auto.
     (*~term t*)
     right; intro J; inversion J; subst. apply wf_tm_gives_term in H6. contradiction.
 Case "x~:bnd_eq t1 t2 t3".
   destruct (term_dec t).
   (*term t*)
   destruct (term_dec t0).
   (*term t0*)
   destruct (type_checking_dec Hwf_sig H0 Hsigex H1 ); auto. 
   destruct H3.
   (*wf_tm t0*)
   destruct (type_checking_dec Hwf_sig H0 Hsigex H2 ); auto. 
   destruct H5.
   (*wf_tm t0*)
   destruct (atomic_type_dec S2 t1).
   (*atomic_type*)
   destruct (value_dec t).
   (*val t*)
   destruct (value_dec t0).
   (*val t0*)
   inversion H3.
   destruct (eq_tm_dec t1 x0).
   (*t1 = x0*)
   inversion H5.
   destruct (eq_tm_dec t1 x1).
   (*t1 = x1*)
   destruct (term_dec t1).
   (*term t1*)
   destruct (type_checking_dec Hwf_sig H0 Hsigex H12 ); auto. 
   destruct H13.
   (*wf_tm t1*)
   inversion H13.
   destruct (eq_tm_dec x2 K_TYPE).
   (*K_TYPE = x2*)
   left. subst. subst. apply wf_env_cons_eq; auto.
   (*K_TYPE <> x2*)
   right; intro J; inversion J; subst.
   assert (K_TYPE = x2).
     eapply wf_tm_unique with (S1:=S1) (S2:=S2) (e:=x0); eauto.
   subst. apply n; auto.
   (*~wf_tm t1*)
   right; intro J; inversion J; subst. apply H13. apply ex_intro with (x:=K_TYPE); auto.
   (*~term t1*)
   right; intro J; inversion J; subst. apply wf_tm_gives_term in H25. contradiction.
   (*t1 != x1*)
   right; intro J; inversion J; subst. 
   assert (x1 = x0).
     eapply wf_tm_unique with (S1:=S1) (S2:=S2) (e:=t0); eauto.
   subst. apply n; auto.
   (*t1 != x0*)
   right; intro J; inversion J; subst. 
   assert (t1 = x0).
     eapply wf_tm_unique with (S1:=S1) (S2:=S2) (e:=t); eauto.
   contradiction.
   (*~val t0*)
   right; intro J; inversion J; subst. contradiction.
   (*~val t*)
   right; intro J; inversion J; subst. contradiction.
   (*~atomic_type*)
   right; intro J; inversion J; subst. contradiction.
   (*~wf_tm t0*)
   right; intro J; inversion J; subst. apply H5. apply ex_intro with (x:=t1); auto.
   (*~wf_tm t*)
   right; intro J; inversion J; subst. apply H3. apply ex_intro with (x:=t1); auto.
   (*~term t0*)
   right; intro J; inversion J; subst. apply wf_tm_gives_term in H10. contradiction.
   (*~term t*)
   right; intro J; inversion J; subst. apply wf_tm_gives_term in H8. contradiction.
 (*~wf_env*)
  right; intro J; inversion J; subst; apply H0; auto. 
Qed.
  


                                 
